from flask import g
from flask.views import MethodView
from flask import current_app as app
from apiflask import abort
from hiddifypanel.auth import login_required
from hiddifypanel.models import *
from hiddifypanel.panel import hiddify
from hiddifypanel.drivers import user_driver



from . import has_permission
from .schema import UserSchema, PostUserSchema, PatchUserSchema, SuccessfulSchema


class UserApi(MethodView):
    decorators = [login_required({Role.super_admin, Role.admin, Role.agent})]

    @app.output(UserSchema)  # type: ignore
    def get(self, uuid):
        """User: Get details of a user"""
        user = User.by_uuid(uuid) or abort(404, "User not found")
        if not has_permission(user):
            abort(403, "You don't have permission to access this user")

        return user.to_schema()  # type: ignore

    @app.input(PatchUserSchema, arg_name="data")  # type: ignore
    @app.output(UserSchema)  # type: ignore
    def patch(self, uuid, data):
        """User: Update a user"""
        user = User.by_uuid(uuid) or abort(404, "user not found")
        if not has_permission(user):
            abort(403, "You don't have permission to access this user")

        for field in User.__table__.columns.keys():  # type: ignore
            if field in ['id', 'expiry_time']:
                continue
            # if field in data:
            #     setattr(user, field, data[field])
        data['old_uuid'] = uuid
        user_driver.remove_client(user)
        dbuser = User.add_or_update(**data) or abort(502, "Unknown issue! User is not patched")
        if dbuser.is_active:
            user_driver.add_client(dbuser)
        # the add_or_update doesn't update the uuid of User, so for now just delete old user after adding new
        # if user.uuid != data['uuid']:
        #     user.remove()
        hiddify.quick_apply_users()
        return dbuser.to_schema()

    @app.output(SuccessfulSchema)  # type: ignore
    def delete(self, uuid):
        """User: Delete a User"""
        user = User.by_uuid(uuid) or abort(404, "user not found")
        if not has_permission(user):
            abort(403, "You don't have permission to access this user")
        user.remove()  # type: ignore
        hiddify.quick_apply_users()
        return {'status': 200, 'msg': 'ok'}



from .user import UserView
from flask import send_from_directory
from flask import Blueprint
from hiddifypanel.database import db

# from .resources import ProductItemResource, ProductResource
from .user import *
from apiflask import APIBlueprint
bp = APIBlueprint("client", __name__, url_prefix="/<proxy_path>/client/", template_folder="templates", enable_openapi=False)


def send_static(path):
    return send_from_directory('static/assets', path)


def init_app(app):
    # @app.route('/<proxy_path>/<user_secret>')
    # @bp.route('/<proxy_path>/<user_secret>/')
    # @app.doc(hide=True)
    # def backward_compatibality():
    #     # TODO: handle none -browser requests
    #     # return redirect(request.url_root.rstrip('/') + f"/{g.proxy_path}/admin/")

    #     return render_template('redirect_to_user.html', user_link=request.url_root.replace('http://', 'https://').rstrip('/') + f"/{g.proxy_path}/#{g.account.name}")

    UserView.register(bp, route_base="/")

    # bp.add_url_rule("/", view_func=index)
    # bp.add_url_rule("/<lang>", view_func=index)
    # bp.add_url_rule("/clash/<meta_or_normal>/<mode>.yml", view_func=clash_config)
    # bp.add_url_rule("/clash/<mode>.yml", view_func=clash_config)
    # bp.add_url_rule("/clash/<meta_or_normal>/proxies.yml", view_func=clash_proxies)
    # bp.add_url_rule("/clash/proxies.yml", view_func=clash_proxies)
    # bp.add_url_rule("/all.txt", view_func=all_configs)
    # bp.add_url_rule('/static/<path:path>',view_func=send_static)
    app.register_blueprint(bp)
    app.register_blueprint(bp, name=f"child_{bp.name}", url_prefix="/<proxy_path>/<int:child_id>/")
    app.register_blueprint(bp, name=f"{bp.name}_uuid", url_prefix="/<proxy_path>/<uuid:secret_uuid>/")
    app.register_blueprint(bp, name=f"child_{bp.name}_uuid", url_prefix="/<proxy_path>/<int:child_id>/<uuid:secret_uuid>/")



from apiflask import Schema, fields, abort
from flask.views import MethodView
from hiddifypanel import hutils
from hiddifypanel.models.role import Role
from flask import current_app as app, make_response, g, request
import os
from ansi2html import Ansi2HTMLConverter
from hiddifypanel.auth import login_required
from hiddifypanel.models import *


class AdminInputLogfileSchema(Schema):
    file = fields.String( metadata={"description": "The log file name"}, required=True)


class AdminLogApi(MethodView):
    @app.input(AdminInputLogfileSchema, arg_name="data", location='form')  # type: ignore
    @app.output(list[fields.String(metadata={"description": "The html of the log"})]) # type: ignore
    @login_required({Role.super_admin})
    def post(self, data):
        """System: View Log file"""
        file_name = data.get('file') or abort(400, "Parameter issue: 'file'")
        log_dir = f"{app.config['HIDDIFY_CONFIG_PATH']}log/system/"
        log_files = hutils.flask.list_dir_files(log_dir)

        file_path = f"{log_dir}{file_name}"
        if file_name not in log_files or not os.path.exists(file_path):
            return abort(404, "Invalid log file")

        logs = ''
        with open(file_path, 'r') as f:
            lines = [line for line in f]
            logs = "".join(lines)

        conv = Ansi2HTMLConverter()
        html_log = f'<div style="background-color:black; color:white;padding:10px">{conv.convert(logs)}</div>'
        resp = make_response(html_log)
        domain = request.args.get("domain")
        resp.headers["Access-Control-Allow-Origin"] = f'*'
        return resp

    def options(self):
        # domain = request.args.get("domain")
        # Domain.query.filter(Domain.domain == domain).first() or abort(404)
        if g.proxy_path != hconfig(ConfigEnum.proxy_path_admin):
            abort(403)
        resp = make_response("")
        resp.headers["Allow"] = "POST"
        resp.headers["Access-Control-Allow-Origin"] = f'*'
        resp.headers["Access-Control-Allow-Headers"] = "Hiddify-API-Key"
        return resp




from flask import g, redirect, request, session
from hiddifypanel.hutils.flask import hurl_for
from flask_login.utils import _get_user
from functools import wraps
from hiddifypanel.models import *
from apiflask import abort as json_abort
from hiddifypanel import hutils
from werkzeug.local import LocalProxy
current_account: "BaseAccount" = LocalProxy(lambda: _get_user())


class AnonymousAccount(BaseAccount):
    __abstract__ = True

    @property
    def uuid(self):
        return None

    @property
    def lang(self) -> Lang | None:
        return None

    @property
    def role(self) -> Role | None:
        return None

    def get_id(self) -> str | None:
        return "-1"

    def __bool__(self):
        return False

    def __eq__(self, other):
        if other is None:
            return True
        if isinstance(other, AnonymousAccount):
            return True
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "AnonymousAccount"


def _get_user():
    if not g.get("__account_store", None):
        g.__account_store = AnonymousAccount()
    return g.__account_store


def admin_session_is_exist():
    return '_admin_id' in session


def logout_user():
    g.__account_store = None
    if '_user_id' in session:
        session.pop('_user_id')
    if '_admin_id' in session:
        session.pop('_admin_id')


def login_user(user: AdminUser | User, remember=False, duration=None, force=False, fresh=True):
    # abort(400, f'logining user: {user} {user.is_active}')
    g.__account_store = user
    # if not user.is_active:
    #     return False

    account_id = user.get_id()  # type: ignore
    # print('account_id', account_id)
    if user.role in {Role.super_admin, Role.admin, Role.agent}:
        session["_admin_id"] = account_id
    else:
        session["_user_id"] = account_id
    # session["_fresh"] = fresh
    # session["_id"] = current_app.login_manager._session_identifier_generator()  # type: ignore

    # if remember:
    #     session["_remember"] = "set"
    #     if duration is not None:
    #         try:
    #             # equal to timedelta.total_seconds() but works with Python 2.6
    #             session["_remember_seconds"] = (
    #                 duration.microseconds
    #                 + (duration.seconds + duration.days * 24 * 3600) * 10**6
    #             ) / 10.0**6
    #         except AttributeError as e:
    #             raise Exception(
    #                 f"duration must be a datetime.timedelta, instead got: {duration}"
    #             ) from e

    # current_app.login_manager._update_request_context_with_user(user)  # type: ignore
    # user_logged_in.send(current_app._get_current_object(), user=_get_user())  # type: ignore
    return True


def login_required(roles: set[Role] | None = None, node_auth: bool = False):

    def decorator(func):
        from flask import has_app_context, current_app
        # Conditionally apply x if has_app_context() is true
        if has_app_context():
            func = current_app.doc(security='Hiddify-API-Key')(func)

        # Always apply y
        func = login_required2(roles, node_auth)(func)
        return func
    return decorator


def login_required2(roles: set[Role] | None = None, node_auth: bool = False):
    '''When both roles and node_auth is set, means authentication can be done by either uuid or unique_id'''

    def wrapper(fn):

        @wraps(fn)
        def decorated_view(*args, **kwargs):
            # print('xxxx', current_account)
            if node_auth and not Child.node and not roles:
                json_abort(403, 'Unauthorized node')
            if not current_account and not node_auth:
                return redirect_to_login()  # type: ignore
            if roles and not Child.node:
                account_role = current_account.role
                if account_role not in roles:
                    return redirect_to_login()  # type: ignore
            return fn(*args, **kwargs)
        return decorated_view
    return wrapper


def get_account_by_api_key(api_key, is_admin):
    # return AdminUser.by_uuid(api_key) if is_admin else User.by_uuid(api_key)
    # api_key equals uuid for now
    return get_account_by_uuid(api_key, is_admin)


def get_account_by_uuid(uuid, is_admin):
    return AdminUser.by_uuid(f'{uuid}') if is_admin else User.by_uuid(f'{uuid}')


def login_by_uuid(uuid,password:str, is_admin: bool)->bool:
    account = get_account_by_uuid(uuid, is_admin)
    if not account:
        return False
    if account.password!=password:
        return False
    return login_user(account, force=True)


def auth_before_request():
    if ".webmanifest" in request.path:
        return

    # print("before_request")
    account = None

    is_admin_path = hutils.flask.is_admin_proxy_path()
    next_url = None

    if g.uuid:
        # print("uuid", g.uuid, is_admin_path)
        account = get_account_by_uuid(g.uuid, is_admin_path)
        # print(account)
        if not account or account.password!="":
            return logout_redirect()
        if is_admin_path:
            next_url = request.url
            next_url = next_url.replace(f'/{g.uuid}/', '/admin/')
            next_url = next_url.replace("/admin/admin/", '/admin/')
            next_url = next_url.replace("http://", "https://")

    elif apikey := request.headers.get("Hiddify-API-Key"):
        account = get_account_by_api_key(apikey, is_admin_path)
        if not account:
            # when parent/child panel needs to call another parent/child api, it will pass its unique id in the header as apikey
            if node := Child.by_unique_id(apikey):
                g.node = node
                return

        if not account:
            return logout_redirect()
    elif request.authorization:
        # print('request.authorization', request.authorization)
        uname = request.authorization.username
        pword = request.authorization.password
        if not pword:
            # print("NO PASSWORD so it is uuid")
            account = get_account_by_uuid(uname, is_admin_path)
        else:
            account = AdminUser.by_username_password(uname, pword) if is_admin_path else User.by_username_password(uname, pword)
        if not account:
            return logout_redirect()

    elif (session_user := session.get('_user_id')) and not is_admin_path:
        # print('session_user', session_user)
        account = User.by_id(int(session_user.split("_")[1]))  # type: ignore
        if not account:
            return logout_redirect()
    elif (session_admin := session.get('_admin_id')) and is_admin_path:
        # print('session_admin', session_admin)
        account = AdminUser.by_id(int(session_admin.split("_")[1]))  # type: ignore
        if not account:
            return logout_redirect()

    if account:
        g.__account_store = account
        # g.account_uuid = account.uuid
        g.is_admin = hutils.flask.is_admin_role(account.role)  # type: ignore
        login_user(account, force=True)
        # print("loggining in")
        if not g.is_admin:
            return
        if next_url is None:
            return
        if not g.user_agent['is_browser']:
            return
        if ".webmanifest" in request.path:
            return

        return redirect(next_url)


def logout_redirect():
    print(f"Incorrect user {current_account}.... loggining out")
    logout_user()
    return redirect_to_login()


def redirect_to_login():
    if hutils.flask.is_api_call(request.path):
        json_abort(403, 'Unathorized')
    # if g.user_agent['is_browser']:
    # return redirect(hurl_for('common_bp.LoginView:basic_0', force=1, next=request.path))
    return redirect(hurl_for('common_bp.LoginView:index', force=1, next=request.path.replace(f'{g.uuid}/',''),user=g.uuid))

    # else:
    #     abort(401, "Unauthorized")
    # return f'/{request.path.split("/")[1]}/?force=1&redirect={request.path}'

# @login_manager.request_loader



import asyncio
import time
from flask import current_app as app, request
from flask import g
from flask.views import MethodView
from apiflask.fields import Dict
from apiflask import Schema
from hiddifypanel.models.usage import DailyUsage
from hiddifypanel.auth import login_required
from hiddifypanel.models import Role, DailyUsage
from hiddifypanel.panel import hiddify, usage
from hiddifypanel import hutils
import json


class UpdateUserUsageApi(MethodView):
    decorators = [login_required({Role.super_admin})]

    def get(self):
        """System: Update User Usage"""
        # time.sleep(5)
        
        return json.dumps(usage.update_local_usage_not_lock(), indent=2)


class AllConfigsApi(MethodView):
    decorators = [login_required({Role.super_admin})]

    def get(self):
        """System: All Configs for configuration"""
        return json.dumps(hiddify.all_configs_for_cli(), indent=2)





class DriverABS:
    def get_all_usage(self): pass
    def get_enabled_users(self): pass
    def add_client(self, user): pass
    def remove_client(self, user): pass

    def is_enabled(self) -> bool: return False



from flask import current_app as app, request
from flask import g
from flask.views import MethodView
from apiflask.fields import Dict
from apiflask import Schema
from hiddifypanel.models.usage import DailyUsage
from hiddifypanel.auth import login_required
from hiddifypanel.models import Role, DailyUsage
from hiddifypanel.panel import hiddify
from hiddifypanel import hutils


class ServerStatusOutputSchema(Schema):
    stats = Dict(required=True,  metadata={"description": "System stats"})
    usage_history = Dict(required=True,  metadata={"description": "System usage history"})


class AdminServerStatusApi(MethodView):
    decorators = [login_required({Role.super_admin, Role.admin, Role.agent})]

    @app.output(ServerStatusOutputSchema)  # type: ignore
    def get(self):
        """System: ServerStatus"""
        dto = ServerStatusOutputSchema()
        dto.stats = {  # type: ignore
            'system': hutils.system.system_stats(),
            'top5': hutils.system.top_processes()
        }
        admin_id = request.args.get("admin_id") or g.account.id
        dto.usage_history = DailyUsage.get_daily_usage_stats(admin_id)  # type: ignore
        return dto



import telebot
from flask import request
from apiflask import abort
from flask_restful import Resource
import time

from hiddifypanel.models import *
from hiddifypanel import Events
from hiddifypanel.cache import cache
logger = telebot.logger


class ExceptionHandler(telebot.ExceptionHandler):
    def handle(self, exception):
        """Improved error handling for Telegram bot exceptions"""
        error_msg = str(exception)
        logger.error(f"Telegram bot error: {error_msg}")
        
        try:
            # Attempt recovery based on error type
            if "webhook" in error_msg.lower():
                if hasattr(bot, 'remove_webhook'):
                    bot.remove_webhook()
                    logger.info("Removed webhook due to error")
            elif "connection" in error_msg.lower():
                # Wait and retry for connection issues
                time.sleep(5)
                return True  # Indicates retry
        except Exception as e:
            logger.error(f"Error during recovery attempt: {str(e)}")
        
        return False  # Don't retry for unknown errors


bot = telebot.TeleBot("1:2", parse_mode="HTML", threaded=False, exception_handler=ExceptionHandler())
bot.username = ''


@cache.cache(1000)
def register_bot_cached(set_hook=False, remove_hook=False):
    return register_bot(set_hook, remove_hook)


def register_bot(set_hook=False, remove_hook=False):
    try:
        global bot
        token = hconfig(ConfigEnum.telegram_bot_token)
        if token:
            bot.token = hconfig(ConfigEnum.telegram_bot_token)
            try:
                bot.username = bot.get_me().username
            except BaseException:
                pass
            if remove_hook:
                bot.remove_webhook()
            domain = Domain.get_panel_link()
            if not domain:
                raise Exception('Cannot get valid domain for setting telegram bot webhook')

            admin_proxy_path = hconfig(ConfigEnum.proxy_path_admin)

            user_secret = AdminUser.get_super_admin_uuid()
            if set_hook:
                bot.set_webhook(url=f"https://{domain}/{admin_proxy_path}/{user_secret}/api/v1/tgbot/")
    except Exception as e:
        logger.error(e)
        


def init_app(app):
    with app.app_context():
        global bot
        token = hconfig(ConfigEnum.telegram_bot_token)
        if token:
            bot.token = token
            try:
                bot.username = bot.get_me().username
            except BaseException:
                pass


class TGBotResource(Resource):
    def post(self):
        try:
            if request.headers.get('content-type') == 'application/json':
                json_string = request.get_data().decode('utf-8')
                update = telebot.types.Update.de_json(json_string)
                bot.process_new_updates([update])
                return ''
            else:
                abort(403)
        except Exception as e:
            print("Error", e)
            import traceback
            traceback.print_exc()
            return "", 500



import re
import flask_babel
import uuid
# from flask_babelex import lazy_gettext as _
from flask import render_template, g, request
from flask_babel import gettext as _
from markupsafe import Markup
import wtforms as wtf
from flask_wtf import FlaskForm
from flask_bootstrap import SwitchField
from hiddifypanel.panel import hiddify
from flask_classful import FlaskView
from wtforms.validators import ValidationError, Length, InputRequired
# from gettext import gettext as _

from hiddifypanel.models import Domain, DomainType, StrConfig, ConfigEnum, get_hconfigs
from hiddifypanel.database import db
from hiddifypanel.auth import login_required
from hiddifypanel import hutils
from hiddifypanel.models import *


class QuickSetup(FlaskView):
    decorators = [login_required({Role.super_admin})]

    def current_form(self, step=None, empty=False, next=False):
        step = int(step or request.form.get("step") or request.args.get('step', "1"))
        if next:
            step = step + 1
        form = {1: get_lang_form,
                2: get_password_form,
                3: get_quick_setup_form,
                4: get_proxy_form}

        return form[step](empty=empty or next)

    def index(self):

        return render_template(
            'quick_setup.html',
            form=self.current_form(),
            # ipv4=hutils.network.get_ip_str(4),
            # ipv6=hutils.network.get_ip_str(6),
            admin_link=admin_link(),
            show_domain_info=True)

    def post(self):
        if request.args.get('changepw') == "true":
            AdminUser.current_admin_or_owner().uuid = str(uuid.uuid4())
            # AdminUser.current_admin_or_owner().password = hutils.random.get_random_password()
            db.session.commit()

        set_hconfig(ConfigEnum.first_setup, False)
        form = self.current_form()
        if not form.validate_on_submit() or form.step.data not in ["1", "2", "3","4"]:
            hutils.flask.flash(_('config.validation-error'), 'danger')
            return render_template(
                'quick_setup.html', form=form,
                admin_link=admin_link(),
                ipv4=hutils.network.get_ip_str(4),
                ipv6=hutils.network.get_ip_str(6),
                show_domain_info=False)
        return form.post(self)


def get_lang_form(empty=False):
    class LangForm(FlaskForm):
        step = wtf.HiddenField(default="1")
        admin_lang = wtf.SelectField(
            _("config.admin_lang.label"), choices=[("en", _("lang.en")), ("fa", _("lang.fa")), ("pt", _("lang.pt")), ("zh", _("lang.zh")), ("ru", _("lang.ru")), ("my", _("lang.my"))],
            description=_("config.admin_lang.description"),
            default=hconfig(ConfigEnum.admin_lang))
        # lang=wtf.SelectField(_("config.lang.label"),choices=[("en",_("lang.en")),("fa",_("lang.fa"))],description=_("config.lang.description"),default=hconfig(ConfigEnum.lang))
        country = wtf.SelectField(
            _("config.country.label"), choices=[("ir", _("Iran")), ("zh", _("China")), ("ru", _("Russia")),  ("other", "Others")],
            description=_("config.country.description"),
            default=hconfig(ConfigEnum.country))
        lang_submit = wtf.SubmitField(_('Submit'))

        def post(self, view):
            set_hconfig(ConfigEnum.lang, self.admin_lang.data)
            set_hconfig(ConfigEnum.admin_lang, self.admin_lang.data)
            set_hconfig(ConfigEnum.country, self.country.data)

            flask_babel.refresh()
            hutils.flask.flash((_('quicksetup.setlang.success')), 'success')

            return render_template(
                'quick_setup.html', form=view.current_form(next=True),
                # admin_link=admin_link(),
                # ipv4=hutils.network.get_ip_str(4),
                # ipv6=hutils.network.get_ip_str(6),
                show_domain_info=False)

    form = LangForm(None)if empty else LangForm()
    form.step.data = "1"
    return form


def get_password_form(empty=False):
    class PasswordForm(FlaskForm):
        step = wtf.HiddenField(default="1")
        admin_pass = wtf.PasswordField(
            _("user.password.title"),
            description=_("user.password.description"),
            default="",validators=[
                
                InputRequired(message=_("user.password.validation-required")), 
                Length(min=8, message=_("user.password.validation-lenght"))
        
            ])
        password_submit = wtf.SubmitField(_('Submit'))

        def post(self, view):
            AdminUser.current_admin_or_owner().update_password(self.admin_pass.data)

            return render_template(
                'quick_setup.html', form=view.current_form(next=True),
                admin_link=admin_link(),
                ipv4=hutils.network.get_ip_str(4),
                ipv6=hutils.network.get_ip_str(6),
                show_domain_info=False)

    form = PasswordForm(None)if empty else PasswordForm()
    form.step.data = "2"
    return form

def get_proxy_form(empty=False):
    class ProxyForm(FlaskForm):
        step = wtf.HiddenField(default="3")
        description_for_fieldset = wtf.TextAreaField("", description=_(f'quicksetup.proxy_cat.description'), render_kw={"class": "d-none"})

        def post(self, view):

            for k, vs in self.data.items():
                ek = ConfigEnum[k]
                if ek != ConfigEnum.not_found:
                    set_hconfig(ek, vs, commit=False)

            db.session.commit()
            # print(cat,vs)
            hutils.proxy.get_proxies.invalidate_all()
            if hutils.node.is_child():
                hutils.node.run_node_op_in_bg(hutils.node.child.sync_with_parent, *[hutils.node.child.SyncFields.hconfigs])

            from .Actions import Actions
            return Actions().reinstall(domain_changed=True)
    boolconfigs = BoolConfig.query.filter(BoolConfig.child_id == Child.current().id).all()

    for cf in boolconfigs:
        if cf.key.category == 'hidden':
            continue
        if cf.key.startswith("sub_") or cf.key.startswith("mux_"):
            continue
        if not cf.key.endswith("_enable") or cf.key in [ConfigEnum.hysteria_obfs_enable, ConfigEnum.tls_padding_enable]:
            continue
        
        field = SwitchField(_(f'config.{cf.key}.label'), default=cf.value, description=_(f'config.{cf.key}.description'))
        setattr(ProxyForm, f'{cf.key}', field)
    setattr(ProxyForm, "submit_global", wtf.fields.SubmitField(_('Submit')))
    form = ProxyForm(None) if empty else ProxyForm()
    form.step.data = "4"
    return form


def get_quick_setup_form(empty=False):
    def get_used_domains():
        configs = get_hconfigs()
        domains = []
        for c in configs:
            if "domain" in c:
                domains.append(configs[c])
        for d in Domain.query.all():
            domains.append(d.domain)
        return domains

    class BasicConfigs(FlaskForm):
        step = wtf.HiddenField(default="2")
        description_for_fieldset = wtf.TextAreaField("", description=_(f'quicksetup.proxy_cat.description'), render_kw={"class": "d-none"})
        domain_regex = "^([A-Za-z0-9\\-\\.]+\\.[a-zA-Z]{2,})$"

        domain_validators = [
            wtf.validators.Regexp(domain_regex, re.IGNORECASE, _("config.Invalid_domain")),
            validate_domain,
            wtf.validators.NoneOf([d.domain.lower() for d in Domain.query.all()], _("config.Domain_already_used")),
            wtf.validators.NoneOf([c.value.lower() for c in StrConfig.query.all() if "fakedomain" in c.key and c.key != ConfigEnum.decoy_domain], _("config.Domain_already_used"))]

        cdn_domain_validators = [
            wtf.validators.Regexp(f'({domain_regex})|(^$)', re.IGNORECASE, _("config.Invalid_domain")),
            validate_domain_cdn,
            wtf.validators.NoneOf([d.domain.lower() for d in Domain.query.all()], _("config.Domain_already_used")),
            wtf.validators.NoneOf([c.value.lower() for c in StrConfig.query.all() if "fakedomain" in c.key and c.key != ConfigEnum.decoy_domain], _("config.Domain_already_used"))]
        domain = wtf.StringField(
            _("domain.domain"),
            domain_validators,
            description=_("domain.description"),
            render_kw={
                "class": "ltr",
                "pattern": domain_validators[0].regex.pattern,
                "title": domain_validators[0].message,
                "required": "",
                "placeholder": "sub.domain.com"})

        cdn_domain = wtf.StringField(
            _("quicksetup.cdn_domain.label"),
            cdn_domain_validators,
            description=_("quicksetup.cdn_domain.description"),
            render_kw={
                "class": "ltr",
                "pattern": domain_validators[0].regex.pattern,
                "title": domain_validators[0].message,
                "placeholder": "sub.domain.com"})
        # enable_telegram = SwitchField(_("config.telegram_enable.label"), description=_("config.telegram_enable.description"), default=hconfig(ConfigEnum.telegram_enable))
        # enable_firewall = SwitchField(_("config.firewall.label"), description=_("config.firewall.description"), default=hconfig(ConfigEnum.firewall))
        block_iran_sites = SwitchField(_("config.block_iran_sites.label"), description=_(
            "config.block_iran_sites.description"), default=hconfig(ConfigEnum.block_iran_sites))
        # enable_vmess = SwitchField(_("config.vmess_enable.label"), description=_("config.vmess_enable.description"), default=hconfig(ConfigEnum.vmess_enable))
        decoy_domain = wtf.StringField(_("config.decoy_domain.label"), description=_("config.decoy_domain.description"), default=hconfig(
            ConfigEnum.decoy_domain), validators=[wtf.validators.Regexp(domain_regex, re.IGNORECASE, _("config.Invalid_domain")), hutils.flask.validate_domain_exist])
        submit = wtf.SubmitField(_('Submit'))

        def post(self, view):
            Domain.query.filter(Domain.domain == f'{hutils.network.get_ip_str(4)}.sslip.io').delete()
            db.session.add(Domain(domain=self.domain.data.lower(), mode=DomainType.direct))
            if self.cdn_domain.data:
                db.session.add(Domain(domain=self.cdn_domain.data.lower(), mode=DomainType.cdn))
            set_hconfig(ConfigEnum.block_iran_sites, self.block_iran_sites.data)
            set_hconfig(ConfigEnum.decoy_domain, self.decoy_domain.data)
            # hiddify.bulk_register_configs([
            #     # {"key": ConfigEnum.telegram_enable, "value": quick_form.enable_telegram.data == True},
            #     # {"key": ConfigEnum.vmess_enable, "value": quick_form.enable_vmess.data == True},
            #     # {"key": ConfigEnum.firewall, "value": quick_form.enable_firewall.data == True},
            #     {"key": ConfigEnum.block_iran_sites, "value": quick_form.block_iran_sites.data == True},
            #     # {"key":ConfigEnum.decoy_domain,"value":quick_form.decoy_domain.data}
            # ])

            return render_template(
                'quick_setup.html', form=view.current_form(next=True),
                # ipv4=hutils.network.get_ip_str(4),
                # ipv6=hutils.network.get_ip_str(6),
                admin_link=admin_link(),
                show_domain_info=False)

    form = BasicConfigs(None) if empty else BasicConfigs()
    form.step.data = "3"
    return form


def validate_domain(form, field):
    domain = field.data
    dip = hutils.network.get_domain_ip(domain)
    if dip is None:
        raise ValidationError(_("Domain can not be resolved! there is a problem in your domain"))

    myips = hutils.network.get_ips()
      # Fixed: Changed from get_ip(4) to get_ip(6)
    if dip not in myips:
        raise ValidationError(_("Domain (%(domain)s)-> IP=%(domain_ip)s is not matched with your ip=%(server_ip)s which is required in direct mode",
                              server_ip=myips, domain_ip=dip, domain=domain))


def validate_domain_cdn(form, field):
    domain = field.data
    if not domain:
        return
    dip = hutils.network.get_domain_ip(domain)
    if dip is None:
        raise ValidationError(_("Domain can not be resolved! there is a problem in your domain"))

    myips = hutils.network.get_ips()
    if dip in myips:
        raise ValidationError(_("In CDN mode, Domain IP=%(domain_ip)s should be different to your ip=%(server_ip)s",
                              server_ip=myips, domain_ip=dip, domain=domain))


def admin_link():
    domains = Domain.get_domains()
    return hiddify.get_account_panel_link(g.account, domains[0] if len(domains)else hutils.network.get_ip_str(4))



import asyncio
from . import test_pb2_grpc 
import sonora.aio
from .test_pb2 import HelloRequest


async def temp():
    async with sonora.aio.insecure_web_channel(f"http://localhost:9000") as channel:
        stub = test_pb2_grpc.HelloStub(channel)
        print(await stub.SayHello(HelloRequest(req="F")))



asyncio.run(temp())


from hiddifypanel.drivers.telemt_api import TelemtApi

from .ssh_liberty_bridge_api import SSHLibertyBridgeApi
from .xray_api import XrayApi
from .singbox_api import SingboxApi
from .wireguard_api import WireguardApi
from hiddifypanel.models import *
from hiddifypanel.panel import hiddify
from collections import defaultdict
from loguru import logger

drivers = [XrayApi(), SingboxApi(), SSHLibertyBridgeApi(), WireguardApi(),TelemtApi()]


def enabled_drivers():
    return [d for d in drivers if d.is_enabled()]


def get_users_usage(reset=True):
    res = {}
    from hiddifypanel.database import db

    # users = db.session.query(User).all()
    # users = list(User.query.all())
    res = defaultdict(lambda: {'usage': 0, 'devices': ''})
    for driver in enabled_drivers():
        try:
            all_usage = driver.get_all_usage()
            for uuid, usage in all_usage.items():
                # print(f"{driver.__class__.__name__} {uuid} usage={usage}")
                if usage:
                    res[uuid]['usage'] += usage
                # res[user]['devices'] +=usage
        except Exception as e:
            print(driver)
            hiddify.error(f'ERROR! {driver.__class__.__name__} has error in update usage {e}')
            logger.exception(f'ERROR! {driver.__class__.__name__} has error in update usage {e}')
    return res


def get_enabled_users():
    from collections import defaultdict
    d = defaultdict(int)
    total = 0
    for driver in enabled_drivers():
        try:
            for u, v in driver.get_enabled_users().items():
                # print(u, "enabled", v, driver)
                if not v:
                    continue
                d[u] += 1
            total += 1
        except Exception as e:
            print(driver)
            hiddify.error(f'ERROR! {driver.__class__.__name__} has error in get_enabled users {e}')
            logger.exception(f'ERROR! {driver.__class__.__name__} has error in get_enabled users {e}')
    # print(d, total)
    res = defaultdict(bool)
    for u, v in d.items():
        # res[u] = v >= total  # ignore singbox
        res[u] = v >= 1
    return res


def add_client(user: User):
    for driver in enabled_drivers():
        try:
            driver.add_client(user)
        except Exception as e:
            hiddify.error(f'ERROR! {driver.__class__.__name__} has error {e} in add client for user={user.uuid} {e}')
            logger.exception(f'ERROR! {driver.__class__.__name__} has error {e} in add client for user={user.uuid} {e}')


def remove_client(user: User):
    for driver in enabled_drivers():
        try:
            driver.remove_client(user)
        except Exception as e:
            hiddify.error(f'ERROR! {driver.__class__.__name__} has error {e} in remove client for user={user.uuid}')
            logger.exception(f'ERROR! {driver.__class__.__name__} has error {e} in remove client for user={user.uuid}')



from flask import render_template, request, redirect, g
from . import fix_flaskadmin_babel
import flask_admin
from flask_admin import Admin
from hiddifypanel import Events
from .DomainAdmin import DomainAdmin
from .AdminstratorAdmin import AdminstratorAdmin


from hiddifypanel.database import db
from hiddifypanel.models import *
from apiflask import APIBlueprint
from flask_adminlte3 import AdminLTE3

flask_bp = APIBlueprint("flask", __name__, template_folder="templates", enable_openapi=False)
admin_bp = APIBlueprint("admin", __name__, template_folder="templates", enable_openapi=False)

flaskadmin = Admin(endpoint="admin", base_template='flaskadmin-layout.html',
                   translations_path="/opt/hiddify-develop/hiddify-panel/src/hiddifypanel/translations/")


def init_app(app):

    from .UserAdmin import UserAdmin
    # admin_secret=StrConfig.query.filter(StrConfig.key==ConfigEnum.admin_secret).first()
    #
    # return
    # admin = Admin(endpoint="admin",index_view=Dashboard(),base_template='lte-master.html',static_url_path="/static/")
    flaskadmin.template_mode = "bootstrap4"
    flaskadmin.init_app(flask_bp)
    adminlte = AdminLTE3()
    adminlte.init_app(app)

    Events.admin_prehook.notify(flaskadmin=flaskadmin, admin_bp=admin_bp)

    @app.route('/<proxy_path>/admin')
    @app.doc(hide=True)
    def auto_route(proxy_path=None, user_secret=None):
        return redirect(request.url.replace("http://", "https://") + "/")

    flaskadmin.add_view(UserAdmin(User, db.session))
    flaskadmin.add_view(DomainAdmin(Domain, db.session))
    flaskadmin.add_view(AdminstratorAdmin(AdminUser, db.session))
    from .NodeAdmin import NodeAdmin
    flaskadmin.add_view(NodeAdmin(Child, db.session))
    from .Dashboard import Dashboard
    from .SettingAdmin import SettingAdmin
    from .commercial_info import CommercialInfo
    from .ProxyAdmin import ProxyAdmin
    from .Actions import Actions
    from .Backup import Backup
    from .QuickSetup import QuickSetup
    Dashboard.register(admin_bp, route_base="/")
    SettingAdmin.register(admin_bp)
    ProxyAdmin.register(admin_bp)
    Actions.register(admin_bp)
    CommercialInfo.register(admin_bp)
    QuickSetup.register(admin_bp)
    Backup.register(admin_bp)

    # admin_bp.add_url_rule('/admin/quicksetup/',endpoint="quicksetup",view_func=QuickSetup.index,methods=["GET"])
    # admin_bp.add_url_rule('/admin/quicksetup/',endpoint="quicksetup-save", view_func=QuickSetup.save,methods=["POST"])

    app.add_url_rule("/<proxy_path>/admin/static/<filename>/", endpoint="admin.static")  # fix bug in admin with blueprint

    flask_bp.debug = True
    app.register_blueprint(admin_bp, url_prefix=f"/<proxy_path>/admin/",)
    app.register_blueprint(admin_bp, name=f'child_{admin_bp.name}', url_prefix=f"/<proxy_path>/<int:child_id>/admin/")
    app.register_blueprint(flask_bp, url_prefix=f"/<proxy_path>/")
    app.register_blueprint(flask_bp, name=f'child_{flask_bp.name}', url_prefix=f"/<proxy_path>/<int:child_id>/")



from flask.views import MethodView

from hiddifypanel.auth import login_required
from hiddifypanel.models import Role
from .schema import PongOutputSchema
from flask import current_app as app


class PingPongApi(MethodView):
    decorators = [login_required(roles={Role.super_admin, Role.admin, Role.agent}, node_auth=True)]

    @app.output(PongOutputSchema)  # type: ignore
    def get(self):
        """Ping Pong: Get"""
        return {'msg': 'GET: PONG'}

    @app.output(PongOutputSchema)  # type: ignore
    def post(self):
        """Ping Pong: Post"""
        return {'msg': 'POST: PONG'}

    @app.output(PongOutputSchema)  # type: ignore
    def patch(self):
        """Ping Pong: Patch"""
        return {'msg': 'PATCH: PONG'}

    @app.output(PongOutputSchema)  # type: ignore
    def delete(self):
        """Ping Pong: Delete"""
        return {'msg': 'DELETE: PONG'}

    @app.output(PongOutputSchema)  # type: ignore
    def put(self):
        """Ping Pong: Put"""
        return {'msg': 'PUT: PONG'}



import datetime


from hiddifypanel.database import db



class Report(db.Model):
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    user_id = db.Column(db.Integer, db.ForeignKey('user.id'), default=0, nullable=False)
    asn_id = db.Column(db.String(200), nullable=False, unique=False)
    city = db.Column(db.String(200))
    country = db.Column(db.String(200))
    latitude = Column(Float,)
    longitude = Column(Float, )
    accuracy_radius = Column(Float, )

    date = db.Column(db.DateTime, nullable=False, default=datetime.datetime.min)
    details = db.relationship('ReportDetail', cascade="all,delete", backref='report', lazy='dynamic',)


class ReportDetail(db.Model):
    report_id = db.Column(db.Integer, db.ForeignKey('report.id'), primary_key=True, )
    proxy_id = db.Column(db.Integer, db.ForeignKey('proxy.id'), primary_key=True, )
    ping = db.Column(db.Integer, default=-1)



from flask_babel import force_locale
from flask_babel import gettext as _
from telebot import types
import datetime

from hiddifypanel.database import db
from hiddifypanel.panel import hiddify
from hiddifypanel.models import *
from hiddifypanel import hutils
from . import bot


@bot.message_handler(commands=['start'], func=lambda message: "admin" in message.text)
def send_welcome(message):
    text = message.text
    # print("dddd",text)
    uuid = text.split()[1].split("_")[1] if len(text.split()) > 1 else None
    if hutils.auth.is_uuid_valid(uuid):
        admin_user = AdminUser.by_uuid(uuid)
        if admin_user:
            admin_user.telegram_id = message.chat.id
            db.session.commit()
    else:
        admin_user = AdminUser.query.filter(AdminUser.telegram_id == message.chat.id).first()

    if admin_user:
        with force_locale(admin_user.lang or hconfig(ConfigEnum.admin_lang)):
            start_admin(message)
        return
    bot.reply_to(message, "error")


def start_admin(message):

    bot.reply_to(message, _("bot.admin_welcome"), reply_markup=admin_keyboard_main())


def get_admin_by_tgid(message):

    tgid = message.chat.id
    return AdminUser.query.filter(AdminUser.telegram_id == tgid).first()


def admin_keyboard_main():

    return types.InlineKeyboardMarkup(keyboard=[[
        types.InlineKeyboardButton(
            text=_("Create Package"),
            callback_data=f'create_package'
        )
    ]
    ]
    )


def admin_keyboard_gig(old_action):
    def keyboard(gig):
        return types.InlineKeyboardButton(
            text=f"{gig} GB",
            callback_data=f"{old_action} {gig}"
        )
    return types.InlineKeyboardMarkup(keyboard=[
        [keyboard(i) for i in range(1, 5)],
        [keyboard(5 * i) for i in range(1, 5)],
        [keyboard(50 * i) for i in range(1, 5)]
    ]
    )


def admin_keyboard_days(old_action):
    def keyboard(days):
        return types.InlineKeyboardButton(
            text=f"{days}",
            callback_data=f"{old_action} {days}"
        )
    return types.InlineKeyboardMarkup(keyboard=[
        [keyboard(i) for i in range(1, 16, 3)],
        [keyboard(30 * i) for i in range(1, 5)]
    ]
    )


def admin_keyboard_count(old_action):
    def keyboard(count):
        return types.InlineKeyboardButton(
            text=f"{count}",
            callback_data=f"{old_action} {count}"
        )
    return types.InlineKeyboardMarkup(keyboard=[
        [keyboard(i) for i in range(1, 5)],
        [keyboard(i * 5) for i in range(1, 5)],
        [keyboard(i * 50) for i in range(1, 5)]
    ]
    )


def admin_keyboard_domain(old_action):
    def keyboard(domain):
        return types.InlineKeyboardButton(
            text=f"{domain.alias or domain.domain}",
            callback_data=f"{old_action} {domain.id}"
        )
    return types.InlineKeyboardMarkup(keyboard=[
        [keyboard(d)]
        for d in Domain.get_domains()
    ]
    )


@bot.callback_query_handler(func=lambda call: call.data.startswith(f'create_package'))
def create_package(call):  # <- passes a CallbackQuery type object to your function
    admin = get_admin_by_tgid(call.message)
    if not (admin):
        return
    with force_locale(admin.lang or hconfig(ConfigEnum.admin_lang)):
        try:
            splt = call.data.split(" ")
            if len(splt) == 1:
                new_text = _("package size?")
                bot.edit_message_text(new_text, call.message.chat.id, call.message.message_id, reply_markup=admin_keyboard_gig(call.data))
                bot.answer_callback_query(call.id, text=_("Ok"), show_alert=False, cache_time=1)
            elif len(splt) == 2:
                new_text = _("package days?")
                bot.edit_message_text(new_text, call.message.chat.id, call.message.message_id, reply_markup=admin_keyboard_days(call.data))
                bot.answer_callback_query(call.id, text=_("Ok"), show_alert=False, cache_time=1)
            elif len(splt) == 3:
                new_text = _("How many?")
                bot.edit_message_text(new_text, call.message.chat.id, call.message.message_id, reply_markup=admin_keyboard_count(call.data))
                bot.answer_callback_query(call.id, text=_("Ok"), show_alert=False, cache_time=1)
            elif len(splt) == 4:
                new_text = _("Domain?")
                bot.edit_message_text(new_text, call.message.chat.id, call.message.message_id, reply_markup=admin_keyboard_domain(call.data))
                bot.answer_callback_query(call.id, text=_("Ok"), show_alert=False, cache_time=1)
            elif len(splt) == 5:
                gig = int(splt[1])
                days = int(splt[2])
                count = int(splt[3])
                domain = int(splt[4])
                new_text = _("Please Wait")
                bot.edit_message_text(new_text, call.message.chat.id, call.message.message_id, reply_markup=None)
                domain = Domain.query.filter(Domain.id == domain).first()
                from . import Usage
                admin_id = admin.id
                admin_name = admin.name
                for i in range(1, count + 1):
                    new_text = _("Please Wait") + f' {i}/{count}'
                    bot.edit_message_text(new_text, call.message.chat.id, call.message.message_id, reply_markup=None)
                    user = User(package_days=days, usage_limit_GB=gig, name=f"{admin_name} auto {i}  {datetime.date.today()}", added_by=admin_id)
                    db.session.add(user)
                    db.session.commit()
                    # bot.send_message(call.message.chat.id,f"Days: {days}     Limit: {gig}GB
                    # #{i}\n\n
                    # https://{domain.domain}/{hconfig(ConfigEnum.proxy_path)}/{user.uuid}/",reply_markup=Usage.user_keyboard(user.uuid))
                    with force_locale(user.lang or hconfig(ConfigEnum.lang)):
                        bot.send_message(call.message.chat.id, Usage.get_usage_msg(user.uuid, domain), reply_markup=Usage.user_keyboard(user.uuid))

                # db.session.commit()
                new_text = _("Finished")
                bot.edit_message_text(new_text, call.message.chat.id, call.message.message_id, reply_markup=admin_keyboard_main())
                bot.answer_callback_query(call.id, text=_("Ok"), show_alert=False, cache_time=1)
                from hiddifypanel.panel import usage
                usage.update_local_usage()
                hiddify.quick_apply_users()

        except Exception as e:
            print(e)
            # import traceback
            # traceback.print_stack()
            # new_text=f"Error {e}"
            # bot.edit_message_text(new_text, call.message.chat.id, call.message.message_id,reply_markup=admin_keyboard_main())
            # bot.answer_callback_query(call.id,cache_time =1)



import os
import sys
from celery import Celery, Task
from celery.schedules import crontab
from dotenv import dotenv_values
from loguru import logger


def init_app(app):
    class FlaskTask(Task):
        def __call__(self, *args: object, **kwargs: object) -> object:
            with app.app_context():
                return self.run(*args, **kwargs)

    celery_app = Celery(app.name, task_cls=FlaskTask)
    
    celery_app.config_from_object(dict(
        broker_url=app.config['REDIS_URI_MAIN'],
        result_backend=app.config['REDIS_URI_MAIN'],
        # task_ignore_result=True,
    ))
    app.extensions["celery"] = celery_app


        # Calls test('hello') every 10 seconds.
    from hiddifypanel.panel import usage
    celery_app.add_periodic_task(60.0, usage.update_local_usage.s(), name='update usage')
    # celery_app.conf.beat_schedule = {
    # 'update_usage': {
    #     'task': 'hiddifypanel.panel.usage.update_local_usage',
    #     'schedule': 30.0, 

    # },
# }
    from hiddifypanel.panel.cli import backup_task
    celery_app.autodiscover_tasks()
    # celery_app.add_periodic_task(30.0, backup_task.s(), name='backup task')
    # celery_app.add_periodic_task(
    #     crontab(hour="*/6", minute=30),
    #     backup_task.delay(),
    # )

    celery_app.add_periodic_task(
        crontab(hour="*/6", minute="0"),
        backup_task.s(),
        name="backup_task "
    )
    
    celery_app.set_default()
    return celery_app



def init_app_no_flask():
    config={}
    for c, v in dotenv_values(os.environ.get("HIDDIFY_CFG_PATH", 'app.cfg')).items():
        if v.isdecimal():
            v = int(v)
        else:
            v = True if v.lower() == "true" else (False if v.lower() == "false" else v)
        config[c] = v
    import hiddifypanel.database 
    hiddifypanel.database.init_no_flask()

    from hiddifypanel.panel import init_db
    while not init_db.is_db_latest():
        logger.error("The database upgrade is required before proceeding. Retrying...")
        import time
        time.sleep(20)
    
    logger.info("Starting background tasks")

    celery_app = Celery()
    
    celery_app.config_from_object(dict(
        broker_url=config['REDIS_URI_MAIN'],
        result_backend=config['REDIS_URI_MAIN'],
        # task_ignore_result=True,
    ))
    

    
        # Calls test('hello') every 10 seconds.
    from hiddifypanel.panel import usage
    celery_app.add_periodic_task(60.0, usage.update_local_usage.s(), name='update usage')
    # celery_app.conf.beat_schedule = {
    # 'update_usage': {
    #     'task': 'hiddifypanel.panel.usage.update_local_usage',
    #     'schedule': 30.0, 

    # },
# }
    from hiddifypanel.panel.cli import backup_task
    celery_app.autodiscover_tasks()
    # celery_app.add_periodic_task(30.0, backup_task.s(), name='backup task')
    # celery_app.add_periodic_task(
    #     crontab(hour="*/6", minute=30),
    #     backup_task.delay(),
    # )

    celery_app.add_periodic_task(
        crontab(hour="*/6", minute="0"),
        # crontab(hour="*", minute="*"),
        backup_task.s(),
        name="backup_task "
    )
    
    celery_app.set_default()
    
    return celery_app


    


import datetime
from enum import auto
from uuid import uuid4
from hiddifypanel.models.role import Role
from dateutil import relativedelta

from strenum import StrEnum
from sqlalchemy import event

from hiddifypanel.database import db
from hiddifypanel.models import Lang
from hiddifypanel.models.base_account import BaseAccount
from hiddifypanel.models.admin import AdminUser


ONE_GIG = 1024 * 1024 * 1024


class UserMode(StrEnum):
    """
    The "UserMode" class is an enumeration that defines five possible modes: "no_reset", "monthly", "weekly",
    "daily", and "disable". These modes represent different settings that can be applied to a user account,
    such as the frequency at which data is reset or whether the account is currently disabled. The class is
    implemented using the "StrEnum" base class and the "auto()" function to generate unique values for each mode.
    """
    no_reset = auto()
    monthly = auto()
    weekly = auto()
    daily = auto()
    # disable = auto()


package_mode_dic = {
    UserMode.daily: 1,
    UserMode.weekly: 7,
    UserMode.monthly: 30
}


class UserDetail(db.Model):
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    user_id = db.Column(db.Integer, db.ForeignKey('user.id'), default=0, nullable=False)
    child_id = db.Column(db.Integer, db.ForeignKey('child.id'), default=0, nullable=False)
    last_online = db.Column(db.DateTime, nullable=False, default=datetime.datetime.min)
    current_usage = db.Column(db.BigInteger, default=0, nullable=False)
    connected_devices = db.Column(db.String(512), default='', nullable=False)

    @property
    def current_usage_GB(self):
        return (self.current_usage or 0) / ONE_GIG

    @current_usage_GB.setter
    def current_usage_GB(self, value):
        self.current_usage = (value or 0) * ONE_GIG

    @property
    def devices(self):
        return []
        # return [] if not self.connected_devices else self.connected_devices.split(",")


class User(BaseAccount):
    """
    This is a model class for a user in a database that includes columns for their ID, UUID, name, online status,
    account expiration date, usage limit, package days, mode, start date, current usage, last reset time, and comment.
    """
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    last_online = db.Column(db.DateTime, nullable=False, default=datetime.datetime.min)
    # removed
    # expiry_time = db.Column(db.Date, default=datetime.date.today() + relativedelta.relativedelta(months=6))
    usage_limit = db.Column(db.BigInteger, default=1000 * ONE_GIG, nullable=False)
    package_days = db.Column(db.Integer, default=90, nullable=False)
    mode = db.Column(db.Enum(UserMode), default=UserMode.no_reset, nullable=False)
    monthly = db.Column(db.Boolean, default=False)  # removed
    start_date = db.Column(db.Date, nullable=True)
    current_usage = db.Column(db.BigInteger, default=0, nullable=False)
    last_reset_time = db.Column(db.Date, default=datetime.date.today())
    added_by = db.Column(db.Integer, db.ForeignKey('admin_user.id'), default=1)
    max_ips = db.Column(db.Integer, default=100, nullable=False)
    details = db.relationship('UserDetail', cascade="all,delete", backref='user', lazy='dynamic',)
    enable = db.Column(db.Boolean, default=True, nullable=False)
    ed25519_private_key = db.Column(db.String(500), default="")
    ed25519_public_key = db.Column(db.String(100), default="")
    wg_pk = db.Column(db.String(50), default="")
    wg_pub = db.Column(db.String(50), default="")
    wg_psk = db.Column(db.String(50), default="")

    @property
    def role(self) -> Role | None:
        return Role.user

    def get_id(self) -> str | None:
        return f'user_{self.id}'

    @property
    def current_usage_GB(self):
        return (self.current_usage or 0) / ONE_GIG

    @current_usage_GB.setter
    def current_usage_GB(self, value):
        self.current_usage = min(1000000 * ONE_GIG, (value or 0) * ONE_GIG)

    @property
    def usage_limit_GB(self):
        return (self.usage_limit or 0) / ONE_GIG

    @usage_limit_GB.setter
    def usage_limit_GB(self, value):
        self.usage_limit = min(1000000 * ONE_GIG, (value or 0) * ONE_GIG)

    @property
    def is_active(self) -> bool:
        """
        The "is_active" function checks if the input user object "user" is active by verifying if their mode is not
        "disable", if their usage limit hasn't been exceeded, and if there are remaining days on their account. The
        function returns a boolean value indicating whether the user is active or not.
        """
        is_active = True
        if not self:
            is_active = False
        elif not self.enable:
            is_active = False
        elif self.usage_limit < self.current_usage:
            is_active = False
        elif self.remaining_days < 0:
            is_active = False
        # elif len(self.devices) > max(3, self.max_ips):
        #     is_active = False
        return is_active

    @property
    def devices(self):
        res = {}
        return res
        for detail in UserDetail.query.filter(UserDetail.user_id == self.id):
            for device in detail.devices:
                res[device] = 1
        return list(res.keys())

    def user_should_reset(self) -> bool:
        # print("start_date",user.start_date, "pack",package_mode_dic.get(user.mode,10000), "total",(datetime.date.today()-user.start_date).days)
        # if user.mode==UserMode.daily:
        #     return 0
        if not self.last_reset_time:
            return True
        elif not self.start_date or (datetime.date.today() - self.last_reset_time).days == 0:
            return False
        return ((datetime.date.today() - self.start_date).days % package_mode_dic.get(self.mode, 10000)) == 0

    def reset_usage(self, commit: bool = False):
        '''Resets the user usages'''
        self.last_reset_time = datetime.date.today()
        self.current_usage_GB = 0

        # there's no usage of UserDetail yet, but we reset it too
        # if ud := UserDetail.query.filter(UserDetail.user_id == self.id).first():
        #    ud.current_usage_GB = 0

        if commit:
            db.session.commit()

    def days_to_reset(self):
        """
        The "days_to_reset" function calculates the number of days until the user's data usage is reset, based on the
        user's start date and mode of usage. The function returns the remaining days as an integer value. If the start
        date is not available, the function returns the total days for the user's mode.
        """
        # print("start_date",user.start_date, "pack",package_mode_dic.get(user.mode,10000), "total",(datetime.date.today()-user.start_date).days)
        # if user.mode==UserMode.daily:
        #     return 0
        if self.start_date:
            days = package_mode_dic.get(self.mode, 10000) - (datetime.date.today() - self.start_date).days % package_mode_dic.get(self.mode, 10000)
        else:
            days = package_mode_dic.get(self.mode, 10000)
        return max(-100000, min(days, 100000))

    @property
    def remaining_days(self) -> int:
        """
        The "remaining_days" function calculates the number of days remaining for a user's account package based on the
        current date and the user's start date. The function returns the remaining days as an integer value. If the start
        date is not available, the function returns the total package days.
        """
        res = -1
        if self.package_days is None:
            res = -1
        elif self.start_date:
            # print(datetime.date.today(), u.start_date,u.package_days, u.package_days - (datetime.date.today() - u.start_date).days)
            res = self.package_days - (datetime.date.today() - self.start_date).days
        else:
            # print("else",u.package_days )
            res = self.package_days
        return min(res, 10000)

    def remove(self, commit=True) -> None:
        from hiddifypanel.drivers import user_driver
        user_driver.remove_client(self)
        db.session.delete(self)
        if commit:
            db.session.commit()

    @classmethod
    def by_uuid(cls, uuid: str, create: bool = False) -> 'User':
        if not isinstance(uuid, str):
            uuid = str(uuid)
        account = User.query.filter(User.uuid == uuid).first()
        if not account and create:
            from hiddifypanel import hutils
            if not hutils.auth.is_uuid_valid(uuid):
                uuid = str(uuid4())

            dbuser = User(uuid=uuid, name="unknown", added_by=AdminUser.current_admin_or_owner().id)
            db.session.add(dbuser)
            db.session.commit()
            account = User.by_uuid(uuid, False)
        return account

    @classmethod
    def remove_by_uuid(cls, uuid: str, commit=True):
        dbuser = User.by_uuid(uuid)
        dbuser.remove(dbuser, commit)

    @classmethod
    def add_or_update(cls, commit: bool = True, **data):
        from hiddifypanel import hutils
        dbuser: User = super().add_or_update(commit=commit, **data)
        if data.get('added_by_uuid'):
            admin = AdminUser.by_uuid(data.get('added_by_uuid'), create=True) or AdminUser.current_admin_or_owner()  # type: ignore
            dbuser.added_by = admin.id
        elif not dbuser.added_by:
            dbuser.added_by = 1

        # if data.get('expiry_time', ''): #v4
        #     last_reset_time = hutils.convert.json_to_time(data.get('last_reset_time', '')) or datetime.date.today()

        #     expiry_time = hutils.convert.json_to_date(data['expiry_time'])
        #     dbuser.start_date = last_reset_time
        #     dbuser.package_days = (expiry_time - last_reset_time).days  # type: ignore
        # el
        if data.get('package_days') is not None:
            dbuser.package_days = data['package_days']

            if data.get('start_date'):
                dbuser.start_date = hutils.convert.json_to_date(data['start_date'])
            elif 'start_date' in data and data['start_date'] is None:
                dbuser.start_date = None

        if (c_GB := data.get('current_usage_GB')) is not None:
            dbuser.current_usage_GB = c_GB
        elif (c := data.get('current_usage')) is not None:
            dbuser.current_usage = c
        elif dbuser.current_usage is None:
            dbuser.current_usage = 0

        if (l_GB := data.get('usage_limit_GB')) is not None:
            dbuser.usage_limit_GB = l_GB
        elif (l := data.get('usage_limit')) is not None:
            dbuser.usage_limit = l
        elif dbuser.usage_limit_GB is None:
            dbuser.usage_limit_GB = 1000

        if data.get('enable') is not None:
            dbuser.enable = data['enable']

        if data.get('ed25519_private_key', '') and data.get('ed25519_public_key', ''):
            dbuser.ed25519_private_key = data.get('ed25519_private_key', '')
            dbuser.ed25519_public_key = data.get('ed25519_public_key', '')
        if data.get('wg_pk') is not None:
            dbuser.wg_pk = data['wg_pk']
        if data.get('wg_pub') is not None:
            dbuser.wg_pub = data['wg_pub']
        if data.get('wg_psk') is not None:
            dbuser.wg_psk = data['wg_psk']

        if data.get('mode') is not None or dbuser.mode is None:
            mode = data.get('mode', UserMode.no_reset)
            if mode == 'disable':
                mode = UserMode.no_reset
                dbuser.enable = False
            dbuser.mode = mode

        if data.get('last_online') is not None:
            dbuser.last_online = hutils.convert.json_to_time(data.get('last_online')) or datetime.datetime.min
        if commit:
            db.session.commit()
        return dbuser

    @staticmethod
    def form_schema(schema):
        return schema.dump(User())

    def to_schema(self):
        user_dict = self.to_dict(dump_id=True)
        from hiddifypanel.panel.commercial.restapi.v2.admin.user_api import UserSchema
        return UserSchema().load(user_dict)

    def to_dict(self, convert_date=True, dump_id=False) -> dict:
        base = super().to_dict()
        from hiddifypanel import hutils
        if dump_id:
            base['id'] = self.id
        if not base.get('lang'):
            from hiddifypanel.models import hconfig, ConfigEnum
            base['lang'] = hconfig(ConfigEnum.lang)
        return {**base,
                'last_online': hutils.convert.time_to_json(self.last_online) if convert_date else self.last_online,
                'usage_limit_GB': self.usage_limit_GB,
                'package_days': self.package_days,
                'mode': self.mode,
                'start_date': hutils.convert.date_to_json(self.start_date)if convert_date else self.start_date,
                'current_usage_GB': self.current_usage_GB,
                'last_reset_time': hutils.convert.time_to_json(self.last_reset_time) if convert_date else self.last_reset_time,
                # 'expiry_time': hutils.convert.date_to_json(self.expiry_time) if convert_date else self.expiry_time,
                'added_by_uuid': self.admin.uuid if self.admin else None,
                'ed25519_private_key': self.ed25519_private_key,
                'ed25519_public_key': self.ed25519_public_key,
                'wg_pk': self.wg_pk,
                'wg_pub': self.wg_pub,
                'wg_psk': self.wg_psk,
                'is_active': self.is_active,
                'enable': self.enable
                }

    # @staticmethod
    # def from_dict(data):
    #     """
    #     Returns a new User object created from a dictionary.
    #     """

    #     return User(
    #         name=data.get('name', ''),
    #         expiry_time=data.get('expiry_time', datetime.date.today() + relativedelta.relativedelta(months=6)),
    #         usage_limit_GB=data.get('usage_limit_GB', 1000),
    #         package_days=data.get('package_days', 90),
    #         mode=UserMode[data.get('mode', 'no_reset')],
    #         monthly=data.get('monthly', False),
    #         start_date=data.get('start_date', None),
    #         current_usage_GB=data.get('current_usage_GB', 0),
    #         last_reset_time=data.get('last_reset_time', datetime.date.today()),
    #         comment=data.get('comment', None),
    #         telegram_id=data.get('telegram_id', None),
    #         added_by=data.get('added_by', 1)
    #     )


@event.listens_for(User, 'before_insert')
def on_user_insert(mapper, connection, target):
    from hiddifypanel import hutils
    hutils.model.gen_username(target)
    # hutils.model.gen_password(target)
    hutils.model.gen_ed25519_keys(target)
    hutils.model.gen_wg_keys(target)



import json
import os

from .abstract_driver import DriverABS
from hiddifypanel.models import User, hconfig, ConfigEnum
from hiddifypanel.panel.run_commander import Command, commander
import redis


USERS_USAGE = "wg:users-usage"


class WireguardApi(DriverABS):
    def get_redis_client(self):
        if not hasattr(self, 'redis_client'):
            self.redis_client = redis.from_url(os.environ.get("REDIS_URI_SSH",""))

        return self.redis_client

    def is_enabled(self) -> bool:
        return hconfig(ConfigEnum.wireguard_enable)

    def __init__(self) -> None:
        super().__init__()
        self.pub_uuid_map={}
    def __load_pubkey_uuid_map(self):
        from hiddifypanel.database import db
        users = db.session.query(User).all()
        self.pub_uuid_map={u.wg_pub: u.uuid for u in users}

    def __convert_pub_key_to_uuid(self,pubkeys):
        res={}
        can_reload_map=True
        for key in pubkeys:
            if uuid:=self.pub_uuid_map.get(key):
                res[key]=uuid
            elif can_reload_map:
                self.__load_pubkey_uuid_map()
                can_reload_map=False
                if uuid:=self.pub_uuid_map.get(key):
                    res[key]=uuid
        return res
            
    def __get_wg_usages(self) -> dict:
        raw_output = commander(Command.update_wg_usage, run_in_background=False)
        data = {}
        for line in raw_output.split('\n'):
            if not line:
                continue
            sections = line.split()
            if len(sections) < 3:
                continue
            data[sections[0]] = {
                'down': int(sections[1]),
                'up': int(sections[2]),
            }
        
        return data

    def __get_local_usage(self) -> dict:
        usage_data = self.get_redis_client() .get(USERS_USAGE)
        if usage_data:
            return json.loads(usage_data)

        return {}

    def __sync_local_usages(self) -> dict:
        local_usage = self.__get_local_usage()
        wg_usage = self.__get_wg_usages()
        
        res = {}
        # remove local usage that is removed from wg usage
        for local_wg_pub in local_usage.copy().keys():
            if local_wg_pub not in wg_usage:
                del local_usage[local_wg_pub]

        
        uuid_map = self.__convert_pub_key_to_uuid(wg_usage.keys())
        for wg_pub, wg_usage in wg_usage.items():
            uuid = uuid_map.get(wg_pub)
            
            if not local_usage.get(wg_pub):
                local_usage[wg_pub] = {"uuid": uuid, "usage": wg_usage}
                continue
            res[uuid] = self.calculate_reset(local_usage[wg_pub]['usage'], wg_usage)
            local_usage[wg_pub] = {"uuid": uuid, "usage": wg_usage}

        self.get_redis_client().set(USERS_USAGE, json.dumps(local_usage))

        return res

    def calculate_reset(self, last_usage: dict, current_usage: dict) -> dict:
        res = {
            'up': current_usage['up'] - last_usage['up'],
            'down': current_usage['down'] - last_usage['down'],
        }

        if res['up'] < 0:
            res['up'] = 0
        if res['down'] < 0:
            res['down'] = 0
        return res

    def get_enabled_users(self):
        if not hconfig(ConfigEnum.wireguard_enable):
            return {}
        usages = self.__get_wg_usages()
        new_wg_pubs = set(usages.keys())
        old_usages = self.__get_local_usage()
        old_wg_pubs = set(old_usages.keys())
        enabled = {u['uuid']: 1 for u in old_usages.values()}
        not_included = new_wg_pubs - old_wg_pubs
        if not_included:
            users = User.query.filter(User.wg_pub.in_(not_included).all())
            for u in users:
                enabled[u.uuid] = 1

        return enabled

    def add_client(self, user):
        pass

    def remove_client(self, user):
        pass

    def get_all_usage(self, reset=True):
        if not hconfig(ConfigEnum.wireguard_enable):
            return {}
        all_usages = self.__sync_local_usages()
        res = {}
        for uuid,use in all_usages.items():
            # if use := all_usages.get(u.wg_pub):
                res[uuid] = use['up'] + use['down']
            # else:
            #     res[u] = 0
        return res



from flask.views import MethodView
from flask import current_app as app

from hiddifypanel.auth import login_required
from hiddifypanel.models import Role
from hiddifypanel import __version__

from .schema import PanelInfoOutputSchema


class PanelInfoApi(MethodView):
    decorators = [login_required(roles={Role.super_admin, Role.admin, Role.agent}, node_auth=True)]

    @app.output(PanelInfoOutputSchema)  # type: ignore
    def get(self):
        res = PanelInfoOutputSchema()
        res.version = __version__  # type: ignore
        return res



import re
import subprocess

from datetime import datetime
from typing import Tuple
from flask import current_app, g
from flask_babel import lazy_gettext as _
from datetime import timedelta
import os
from hiddifypanel.cache import cache
from hiddifypanel.models import *
from hiddifypanel.database import db
from hiddifypanel.hutils.utils import *
from hiddifypanel import hutils
from hiddifypanel.panel.run_commander import commander, Command
import subprocess
to_gig_d = 1000 * 1000 * 1000


# def add_temporary_access():
#     random_port = random.randint(30000, 50000)
#     # exec_command(
#     #     f'sudo /opt/hiddify-manager/hiddify-panel/temporary_access.sh {random_port} &')

#     # run temporary_access.sh
#     commander(Command.temporary_access, port=random_port)
#     temp_admin_link = f"http://{hutils.network.get_ip_str(4)}:{random_port}{get_admin_path()}"
#     g.temp_admin_link = temp_admin_link


# with user panel url format we don't really need this function
def add_short_link(link: str, period_min: int = 5) -> Tuple[str, int]:
    short_code, expire_date = add_short_link_imp(link, period_min)
    return short_code, (expire_date - datetime.now()).seconds


@cache.cache(ttl=300)
# TODO: Change ttl dynamically
def add_short_link_imp(link: str, period_min: int = 5) -> Tuple[str, datetime]:
    # pattern = "\^/([^/]+)(/)?\?\$\ {return 302 " + re.escape(link) + ";}"

    pattern = r"([^/]+)\("

    with open(os.environ['HIDDIFY_CONFIG_PATH'] + "/nginx/parts/short-link.conf", 'r') as f:
        for line in f:
            if link in line:
                return re.search(pattern, line).group(1), datetime.now() + timedelta(minutes=period_min)

    short_code = hutils.random.get_random_string(6, 10).lower()
    # exec_command(
    #     f'sudo /opt/hiddify-manager/nginx/add2shortlink.sh {link} {short_code} {period_min} &')

    commander(Command.temporary_short_link, url=link, slug=short_code, period=period_min)

    return short_code, datetime.now() + timedelta(minutes=period_min)


def exec_command(cmd, cwd=None):
    try:
        subprocess.Popen(cmd.split(" "))  # run in background
    except Exception as e:
        print(e)


def quick_apply_users():
    # run install.sh apply_users
    commander(Command.apply_users)

    return {"status": 'success'}


# Importing socket library

# Function to display hostname and
# IP address


def get_html_user_link(model: BaseAccount, domain: Domain):
    is_cdn = domain.mode == DomainType.cdn if isinstance(domain, Domain) else False
    res = ""
    d = domain.domain
    if "*" in d:
        d = d.replace("*", hutils.random.get_random_string(5, 15))

    # for showing child/node link (we send child_id to get_account_panel_link to get domain proxy path correctly)
    d_child_id = domain.child_id
    link = f'{get_account_panel_link(model, d,child_id=d_child_id)}#{hutils.encode.unicode_slug(model.name)}'

    text = domain.alias or domain.domain
    color_cls = 'info'

    if isinstance(domain, Domain) and not domain.sub_link_only and domain.mode in [DomainType.cdn, DomainType.auto_cdn_ip]:
        auto_cdn = (domain.mode == DomainType.auto_cdn_ip) or (domain.cdn_ip and "MTN" in domain.cdn_ip)
        color_cls = "success" if auto_cdn else 'warning'
        text = f'<span class="badge badge-secondary" >{"Auto" if auto_cdn else "CDN"}</span> ' + text

    res += f"<a target='_blank' data-copy='{link}' href='{link}' class='btn btn-xs btn-{color_cls} ltr share-link' ><i class='fa-solid fa-arrow-up-right-from-square d-none'></i> {text}</a>"

    return res


def reinstall_action(complete_install=False, domain_changed=False, do_update=False):
    from hiddifypanel.panel.admin.Actions import Actions
    action = Actions()
    if do_update:
        return action.update()
    return action.reinstall(complete_install=complete_install, domain_changed=domain_changed)


def check_need_reset(old_configs, do=False):
    restart_mode = ApplyMode.nothing
    for c in old_configs:
        if c.apply_mode == ApplyMode.nothing:
            continue
        # c=ConfigEnum(c)
        if restart_mode == ApplyMode.reinstall:
            break
        if old_configs[c] != hconfig(c):
            restart_mode = c.apply_mode
    if old_configs[ConfigEnum.proxy_path_admin] != hconfig(ConfigEnum.proxy_path_admin):
        g.new_proxy_path = hconfig(ConfigEnum.proxy_path_admin)
        g.force_proxy_path = g.proxy_path
    # do_full_install=old_config[ConfigEnum.telegram_lib]!=hconfig(ConfigEnum.telegram_lib)
    if old_configs[ConfigEnum.package_mode] != hconfig(ConfigEnum.package_mode):
        return reinstall_action(do_update=True)
    if not (do and restart_mode == ApplyMode.reinstall):
        return hutils.flask.flash_config_success(restart_mode=restart_mode, domain_changed=False)

    return reinstall_action(complete_install=True, domain_changed=False)


def get_child(unique_id):
    child_id = Child.current().id
    if unique_id is None or unique_id in ["self", "default", str(hconfig(ConfigEnum.unique_id))]:
        child_id = 0
    else:
        child = Child.query.filter(Child.unique_id == str(unique_id)).first()
        # TODO: this doesn't work because name and mode fields are nullable
        if not child:
            child = Child(unique_id=str(unique_id))
            db.session.add(child)
            db.session.commit()
            child = Child.query.filter(Child.unique_id == str(unique_id)).first()
        child_id = child.id
    return child_id


def dump_db_to_dict():
    return {"childs": [u.to_dict() for u in db.session.query(Child).all()],
            "users": [u.to_dict() for u in db.session.query(User).all()],
            "domains": [u.to_dict() for u in db.session.query(Domain).all()],
            "proxies": [u.to_dict() for u in db.session.query(Proxy).all()],
            # "parent_domains": [] if not hconfig(ConfigEnum.license) else [u.to_dict() for u in ParentDomain.query.all()],
            'admin_users': [d.to_dict() for d in db.session.query(AdminUser).all()],
            "hconfigs": [*[u.to_dict() for u in db.session.query(BoolConfig).all()],
                         *[u.to_dict() for u in db.session.query(StrConfig).all()]]
            }


def get_ids_without_parent(input_dict):
    selector = "uuid"
    # Get all parent_uuids in a set for faster lookup
    parent_uuids = {item.get(f'parent_admin_uuid') for item in input_dict.values()
                    if item.get(f'parent_admin_uuid') is not None
                    and item.get(f'parent_admin_uuid') != item.get('uuid')}
    print("PARENTS", parent_uuids)
    uuids = {v['uuid']: v for v in input_dict.values()}
    # Find all uuids that do not have a parent_uuid in the dict
    uuids_without_parent = [key for key, item in input_dict.items()
                            if item.get(f'parent_admin_uuid') is None
                            or item.get(f'parent_admin_uuid') == item.get('uuid')
                            or item[f'parent_admin_uuid'] not in uuids]
    print("abondon uuids", uuids_without_parent)
    return uuids_without_parent


def set_db_from_json(json_data, override_child_unique_id=True, set_users=True, set_domains=True, set_proxies=True, set_settings=True, remove_domains=False, remove_users=False,
                     override_unique_id=True, set_admins=True, override_root_admin=False, replace_owner_admin=False, fix_admin_hierarchy=True, set_child=True):
    new_rows = []

    # override root child unique id
    if override_child_unique_id:
        backup_child_unique_id = get_backup_child_unique_id(json_data)
        replace_backup_child_unique_id(json_data, backup_child_unique_id, Child.current().unique_id)

    # restore childs
    if set_child and 'childs' in json_data:
        Child.bulk_register(json_data['childs'], commit=True)

    uuids_without_parent = get_ids_without_parent({u['uuid']: u for u in json_data['admin_users']})
    print('uuids_without_parent===============', uuids_without_parent)
    if replace_owner_admin and len(uuids_without_parent):
        new_owner_uuid = uuids_without_parent[0]
        old_owner = AdminUser.query.filter(AdminUser.id == 1).first()
        old_uuid_admin = AdminUser.query.filter(AdminUser.uuid == new_owner_uuid).first()
        if old_owner and not old_uuid_admin:
            old_owner.uuid = new_owner_uuid
            db.session.commit()

    all_admins = {u.uuid: u for u in AdminUser.query.all()}
    uuids_without_parent = [uuid for uuid in uuids_without_parent if uuid not in all_admins]
    print('uuids_not admin exist===============', uuids_without_parent)

    if "admin_users" in json_data:
        for u in json_data['admin_users']:
            if override_root_admin and u['uuid'] in uuids_without_parent:
                u['uuid'] = AdminUser.current_admin_or_owner().uuid
            if u['parent_admin_uuid'] in uuids_without_parent:
                u['parent_admin_uuid'] = AdminUser.current_admin_or_owner().uuid
        # fix admins hierarchy
        if fix_admin_hierarchy and len(json_data['admin_users']) > 2:
            hierarchy_is_ok = False
            for u in json_data['admin_users']:
                if u['uuid'] == AdminUser.current_admin_or_owner().uuid:
                    continue
                if u['parent_admin_uuid'] == AdminUser.current_admin_or_owner().uuid:
                    hierarchy_is_ok = True
                    break
            if not hierarchy_is_ok:
                json_data['admin_users'][1]['parent_admin_uuid'] = AdminUser.current_admin_or_owner().uuid

    if "users" in json_data and override_root_admin:
        for u in json_data['users']:
            if u['added_by_uuid'] in uuids_without_parent:
                u['added_by_uuid'] = AdminUser.current_admin_or_owner().uuid

    if set_admins and 'admin_users' in json_data:
        AdminUser.bulk_register(json_data['admin_users'], commit=True)
    if set_users and 'users' in json_data:
        User.bulk_register(json_data['users'], commit=False, remove=remove_users)
    if set_domains and 'domains' in json_data:
        Domain.bulk_register(json_data['domains'], commit=False, remove=remove_domains)

    if set_settings and 'hconfigs' in json_data:
        bulk_register_configs(json_data["hconfigs"], commit=True, override_unique_id=override_unique_id)
        if 'proxies' in json_data:
            Proxy.bulk_register(json_data['proxies'], commit=False)

    ids_without_parent = get_ids_without_parent({u.id: u.to_dict() for u in AdminUser.query.all()})
    owner = AdminUser.get_super_admin()
    ids_without_parent = [id for id in ids_without_parent if id != owner.id]

    for u in AdminUser.query.all():
        if u.parent_admin_id in ids_without_parent:
            u.parent_admin_id = owner.id
    # for u in User.query.all():
    #     if u.added_by in uuids_without_parent:
    #         u.added_by = g.account.id

    db.session.commit()


def get_domain_btn_link(domain):
    text = domain.alias or domain.domain
    color_cls = "info"
    if domain.mode in [DomainType.cdn, DomainType.auto_cdn_ip]:
        auto_cdn = (domain.mode == DomainType.auto_cdn_ip) or (domain.cdn_ip and "MTN" in domain.cdn_ip)
        color_cls = "success" if auto_cdn else 'warning'
        text = f'<span class="badge badge-secondary" >{"Auto" if auto_cdn else "CDN"}</span> ' + text
    res = f"<a target='_blank' href='#' class='btn btn-xs btn-{color_cls} ltr' ><i class='fa-solid fa-arrow-up-right-from-square d-none'></i> {text}</a>"
    return res


def get_ssh_client_version(user):
    return 'SSH-2.0-OpenSSH_7.4p1'

def is_fake_domain(model:Domain):
        return model.mode in {DomainType.fake,DomainType.reality,DomainType.special_reality_tcp,DomainType.special_reality_grpc,DomainType.special_reality_xhttp}

def get_account_panel_link(account: BaseAccount, host: str, is_https: bool = True, prefere_path_only: bool = False, child_id=None):
    if child_id is None:
        child_id = Child.current().id
    is_admin = isinstance(account, AdminUser)
    basic_auth = False  # is_admin #because safri does not support it.

    link = ""
    if basic_auth or not prefere_path_only:
        link = "https://" if is_https else "http://"
        if basic_auth:
            link += f'{account.uuid}@'
        link += str(host)
    proxy_path = hconfig(ConfigEnum.proxy_path_admin if is_admin else ConfigEnum.proxy_path_client, child_id)
    link += f'/{proxy_path}/'

    # if child_id != 0:
    #     child = Child.by_id(child_id)
    #     link += f"{child.id}/"

    if basic_auth:
        link += "l"
    else:
        link += f'{account.uuid}/'
    return link


def is_telegram_proxy_enable(domains=None) -> bool:
    if not hconfig(ConfigEnum.telegram_enable):
        return False

    valid_domain_types = [DomainType.direct, DomainType.relay, DomainType.old_xtls_direct]
    res = False
    if domains:
        res = any(d.mode in valid_domain_types for d in domains)
    else:
        res = True if Domain.query.filter(Domain.mode.in_(valid_domain_types)).first() else False

    return res


def clone_model(model):
    """Clone an arbitrary sqlalchemy model object without its primary key values."""
    # Ensure the model’s data is loaded before copying.
    # model.id
    new_model = model.__class__()
    table = model.__table__
    for k in table.columns.keys():
        if k == "id":
            continue
        setattr(new_model, f'{k}', getattr(model, k))

    return new_model


def replace_backup_child_unique_id(backupdata: dict, old_child_unique_id: str, new_child_unique_id: str):
    for k, v in backupdata.copy().items():
        if k == 'admin_users' or k == 'users':
            continue
        if k == 'childs':
            if len(v) < 1:
                continue

            if v[0]['unique_id'] == old_child_unique_id or v[0]['unique_id'] == 'self' or v[0]['unique_id'] == 'default':
                v[0]['unique_id'] = new_child_unique_id
        else:
            for item in v:
                if item['child_unique_id'] == old_child_unique_id or item['child_unique_id'] == 'self' or item['child_unique_id'] == 'default':
                    item['child_unique_id'] = new_child_unique_id


def get_backup_child_unique_id(backupdata: dict) -> str:
    if len(backupdata.get('childs', [])) == 0:
        return "self"
    return backupdata['childs'][0]['unique_id']


def all_configs_for_cli():
    valid_users = [u.to_dict(dump_id=True) for u in User.query.filter((User.usage_limit > User.current_usage)).all() if u.is_active]
    host_child_ids = [c.id for c in Child.query.filter(Child.mode == ChildMode.virtual).all()]
    domains = Domain.query.filter(Domain.child_id.in_(host_child_ids),~Domain.domain.contains("*")).all()
    configs = {
        "users": valid_users,
        "domains": [u.to_dict(dump_ports=True, dump_child_id=True) for u in domains],
        # "hconfigs": get_hconfigs(json=True),
        "chconfigs": get_hconfigs_childs(host_child_ids, json=True)
    }

    def_user = User.query.filter(User.name == 'default').first() if User.query.count() == 1 else None
    

    configs['chconfigs'][0]['first_setup'] = def_user is not None and Domain.query.filter(Domain.domain.contains("sslip.io")).limit(1).count() > 0
    server_ip = hutils.network.get_ip_str(4)
    owner = AdminUser.get_super_admin()
    configs['api_key'] = owner.uuid
    configs['api_path'] = hconfig(ConfigEnum.proxy_path_admin)
    configs['admin_path'] = get_account_panel_link(owner, server_ip, is_https=False, prefere_path_only=True)
    configs['panel_links'] = []
    configs['panel_links'].append(get_account_panel_link(owner, server_ip, is_https=False))
    configs['panel_links'].append(get_account_panel_link(owner, server_ip))
    
    domains = Domain.get_domains()
    for d in domains:
        configs['panel_links'].append(get_account_panel_link(owner, d.domain))
    

    return configs



from hiddifypanel.auth import login_required

from wtforms.validators import Regexp
from hiddifypanel.models import *
from wtforms.validators import Regexp, ValidationError
from .adminlte import AdminLTEModelView
from flask_babel import lazy_gettext as _
from wtforms.validators import Regexp
from flask_babel import gettext as __
from flask import request  # type: ignore
from markupsafe import Markup

from flask import g
import datetime
from wtforms import PasswordField, SelectField

from hiddifypanel.panel import hiddify
from hiddifypanel import hutils


class AdminModeField(SelectField):
    def __init__(self, label=None, validators=None, **kwargs):
        super(AdminModeField, self).__init__(label, validators, **kwargs)
        if g.account.mode in [AdminMode.agent, AdminMode.admin]:
            self.choices = [(AdminMode.agent.value, 'agent')]
        elif g.account.mode == AdminMode.admin:
            self.choices = [(AdminMode.agent.value, 'agent'), (AdminMode.admin.value, 'Admin'),]
        elif g.account.mode == AdminMode.super_admin:
            self.choices = [(AdminMode.agent.value, 'agent'), (AdminMode.admin.value, 'Admin'), (AdminMode.super_admin.value, 'Super Admin')]


class SubAdminsField(SelectField):
    def __init__(self, label=None, validators=None, *args, **kwargs):
        kwargs.pop("allow_blank")
        super().__init__(label, validators, *args, **kwargs)
        self.choices = [(admin.id, admin.name) for admin in g.account.sub_admins]
        self.choices += [(g.account.id, g.account.name)]


class AdminstratorAdmin(AdminLTEModelView):
    column_hide_backrefs = False
    column_list = ["name", 'UserLinks', 'mode', 'can_add_admin', 'max_active_users', 'max_users', 'online_users', 'comment',]
    form_columns = ["name", 'mode', 'can_add_admin', 'max_active_users', 'max_users', 'comment', "uuid","new_password"]
    list_template = 'model/admin_list.html'
    # column_editable_list = ['name']
    # edit_modal = True
    # form_overrides = {'work_with': Select2Field}

    form_overrides = {
        'mode': AdminModeField,
        'parent_admin': SubAdminsField
    }
    form_extra_fields = {
        'new_password': PasswordField('New Password',description="If empty, no change")
    }
    column_labels = {
        "Actions": _("actions"),
        "UserLinks": _("user.user_links"),
        "name": _("user.name"),
        "mode": _("Mode"),
        "uuid": _("user.UUID"),
        "comment": _("Note"),
        'max_active_users': _("Max Active Users"),
        'max_users': _('Max Users'),
        "password":_("user.password.title"),
        "online_users": _("Online Users"),
        'can_add_admin': _("Can add sub admin")

    }
    form_args = {
        'uuid': {
            'validators': [Regexp(r'^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$', message=__("Should be a valid uuid"))]
            #     'label': 'First Name',
            #     'validators': [required()]
        }}

    column_descriptions = dict(
        comment=_("Add some text that is only visible to super_admin."),
        mode=_("admin.define_mode"),
    )
    # create_modal = True
    can_export = False

    # column_list = ["domain",'sub_link_only', "mode","alias", "domain_ip", "cdn_ip"]
    # column_editable_list=["domain"]
    # column_filters=["domain","mode"]

    column_searchable_list = ["name", "uuid"]

    # form_columns=['domain','sub_link_only','alias','mode','cdn_ip','show_domains']

    def _ul_formatter(view, context, model, name):

        return Markup(" ".join([hiddify.get_html_user_link(model, d) for d in Domain.get_domains()]))

    @property
    def can_create(self):
        return g.account.can_add_admin or g.account.mode == AdminMode.super_admin

    def _name_formatter(view, context, model, name):

        d = request.host
        if d:

            href = hiddify.get_account_panel_link(model, d) + f'#{hutils.encode.url_encode(model.name)}'
            link = f"<a target='_blank' class='share-link' data-copy='{href}' href='{href}'>{model.name} <i class='fa-solid fa-arrow-up-right-from-square'></i></a>"
            if model.parent_admin:
                return Markup(model.parent_admin.name + "&rlm;&lrm; / &rlm;&lrm;" + link)
            return Markup(link)
        else:
            return model.name

    def _online_users_formatter(view, context, model, name):
        last_day = datetime.datetime.now() - datetime.timedelta(days=1)
        u = model.recursive_users_query().filter(User.last_online > last_day).count()
        t = model.recursive_users_query().count()
        # actives=[u for u in model.recursive_users_query().all() if u.is_active]
        # allusers=model.recursive_users_query().count()
        # onlines=[p for p in  users  if p.last_online and p.last_online>last_day]
        # return Markup(f"<a class='btn btn-xs btn-default' href='{hurl_for('flask.user.index_view',admin_id=model.id)}'> {_('Online')}: {onlines}</a>")
        rate = round(u * 100 / (t + 0.000001))
        state = "danger" if u >= t else ('warning' if rate > 80 else 'success')
        color = "#ff7e7e" if u >= t else ('#ffc107' if rate > 80 else '#9ee150')
        return Markup(f"""
        <div class="progress progress-lg position-relative" style="min-width: 100px;">
          <div class="progress-bar progress-bar-striped" role="progressbar" style="width: {rate}%;background-color: {color};" aria-valuenow="{rate}" aria-valuemin="0" aria-valuemax="100"></div>
              <span class='badge position-absolute' style="left:auto;right:auto;width: 100%;font-size:1em">{u} {_('user.home.usage.from')} {t}</span>

        </div>
        """)

    def _max_users_formatter(view, context, model, name):
        u = model.recursive_users_query().count()
        if model.mode == AdminMode.super_admin:
            return f"{u} / ∞"
        t = model.max_users
        rate = round(u * 100 / (t + 0.000001))
        state = "danger" if u >= t else ('warning' if rate > 80 else 'success')
        color = "#ff7e7e" if u >= t else ('#ffc107' if rate > 80 else '#9ee150')
        return Markup(f"""
        <div class="progress progress-lg position-relative" style="min-width: 100px;">
          <div class="progress-bar progress-bar-striped" role="progressbar" style="width: {rate}%;background-color: {color};" aria-valuenow="{rate}" aria-valuemin="0" aria-valuemax="100"></div>
              <span class='badge position-absolute' style="left:auto;right:auto;width: 100%;font-size:1em">{u} {_('user.home.usage.from')} {t}</span>

        </div>
        """)

    def _max_active_users_formatter(view, context, model, name):
        # `User.is_active` is a python property, not a SQL column.
        # So we must evaluate it on model instances instead of using query.filter.
        active_count = sum(1 for user in model.recursive_users_query().all() if user.is_active)
        if model.mode == AdminMode.super_admin:
            return f"{active_count} / ∞"
        t = model.max_active_users
        rate = round(active_count * 100 / (t + 0.000001))
        color = "#ff7e7e" if active_count >= t else ('#ffc107' if rate > 80 else '#9ee150')
        
        return Markup(f"""
        <div class="progress progress-lg position-relative" style="min-width: 100px;">
          <div class="progress-bar progress-bar-striped" role="progressbar" style="width: {rate}%;background-color: {color};" aria-valuenow="{rate}" aria-valuemin="0" aria-valuemax="100"></div>
              <span class='badge position-absolute' style="left:auto;right:auto;width: 100%;font-size:1em">{active_count} {_('user.home.usage.from')} {t}</span>

        </div>
        """)

    column_formatters = {
        'name': _name_formatter,
        'online_users': _online_users_formatter,
        'max_users': _max_users_formatter,
        'max_active_users': _max_active_users_formatter,
        'UserLinks': _ul_formatter

    }

    def search_placeholder(self):
        return f"{_('search')} {_('user.UUID')} {_('user.name')}"

    # @login_required(roles={Role.super_admin, Role.admin})
    def is_accessible(self):
        if login_required(roles={Role.super_admin, Role.admin, Role.agent})(lambda: True)() != True:
            return False
        return True

    def get_query(self):
        # Get the base query
        query = super().get_query()

        admin_ids = g.account.recursive_sub_admins_ids()
        query = query.filter(AdminUser.id.in_(admin_ids))

        return query

    # Override get_count_query() to include the filter condition in the count query
    def get_count_query(self):
        # Get the base count query
        query = super().get_count_query()

        admin_ids = g.account.recursive_sub_admins_ids()
        query = query.filter(AdminUser.id.in_(admin_ids))

        return query

    def on_model_change(self, form, model, is_created):

        # if model.id==1:
        #     model.parent_admin_id=0
        #     model.parent_admin=None
        # else:
        #     model.parent_admin_id=1
        #     model.parent_admin=AdminUser.query.filter(AdminUser.id==1).first()
        
        if model.id != 1 and model.parent_admin is None:
            model.parent_admin_id = g.account.id
            model.parent_admin = g.account

        if g.account.mode != AdminMode.super_admin and model.mode == AdminMode.super_admin:
            raise ValidationError("Sub-Admin can not have more power!!!!")
        if g.account.mode == AdminMode.agent and model.mode != AdminMode.agent:
            raise ValidationError("Sub-Admin can not have more power!!!!")
        
        if not model.new_password and is_created:
            raise ValidationError("Password for new admin is needed.")
        if model.new_password:
            model.password=model.new_password


    def on_model_delete(self, model):
        model.remove()

    def on_form_prefill(self, form, id=None):
        if g.account.mode != AdminMode.super_admin:
            del form.mode
            del form.can_add_admin

        if g.account.id == form._obj.id:
            del form.max_users
            del form.max_active_users
            del form.comment
            del form.can_add_admin
            if getattr(form, 'mode'):
                del form.mode
        elif form._obj.mode == AdminMode.super_admin:
            del form.max_users
            del form.max_active_users
            del form.can_add_admin

    def after_model_change(self, form, model, is_created):
        if hutils.node.is_parent():
            hutils.node.run_node_op_in_bg(hutils.node.parent.request_childs_to_sync)

    def after_model_delete(self, model):
        if hutils.node.is_parent():
            hutils.node.run_node_op_in_bg(hutils.node.parent.request_childs_to_sync)



import grpc

# from hiddifypanel.hasync.config import BoolConfig 
from hiddifypanel.models.config import hconfig
from hiddifypanel.models.config_enum import ConfigEnum
from . import test_pb2
from . import test_pb2_grpc

class HelloServicer(test_pb2_grpc.HelloServicer):
    """Missing associated documentation comment in .proto file."""

    def SayHello(self, request: test_pb2.HelloRequest, context) -> test_pb2.HelloResponse:
            
            return test_pb2.HelloResponse(res=hconfig(ConfigEnum.log_level))



from apiflask import fields, Schema
from marshmallow import ValidationError

from hiddifypanel.models import DomainType, ProxyProto, ProxyL3, ProxyTransport, ProxyCDN, ConfigEnum, ChildMode
from hiddifypanel.panel.commercial.restapi.v2.admin.schema import UserSchema, AdminSchema


def hconfig_key_validator(value):
    if value not in [c.name for c in ConfigEnum]:
        raise ValidationError(f"{value} is not a valid hconfig key.")
    return value


class DomainSchema(Schema):
    child_unique_id = fields.String( metadata={"description": "The child's unique id"})
    domain = fields.String(required=True,  metadata={"description": "The domain name"})
    alias = fields.String( metadata={"description": "The domain alias"}, allow_none=True)
    sub_link_only = fields.Boolean(required=True,  metadata={"description": "Is the domain sub link only"})
    mode = fields.Enum(DomainType, required=True,  metadata={"description": "The domain type"})
    cdn_ip = fields.String( metadata={"description": "The cdn ip"}, allow_none=True)
    grpc = fields.Boolean(required=True,  metadata={"description": "Is the domain grpc"})
    servernames = fields.String( metadata={"description": "The servernames"}, allow_none=True)
    show_domains = fields.List(fields.String(),  metadata={"description":"The list of domains to show"})


class ProxySchema(Schema):
    child_unique_id = fields.String( metadata={"description": "The child's unique id"})
    name = fields.String(required=True,  metadata={"description": "The proxy name"})
    enable = fields.Boolean(required=True,  metadata={"description": "Is the proxy enabled"})
    proto = fields.Enum(ProxyProto, required=True,  metadata={"description": "The proxy protocol"})
    l3 = fields.Enum(ProxyL3, required=True,  metadata={"description": "The proxy l3"})
    transport = fields.Enum(ProxyTransport, required=True,  metadata={"description": "The proxy transport"})
    cdn = fields.Enum(ProxyCDN, required=True,  metadata={"description": "The proxy cdn"})


class StringOrBooleanField(fields.Field):
    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, (str, bool)):
            return str(value)
        else:
            raise ValidationError("Value must be a string or a boolean.")

    def _serialize(self, value, attr, obj, **kwargs):
        return value


class HConfigSchema(Schema):
    child_unique_id = fields.String( metadata={"description": "The child's unique id"})
    key = fields.String(required=True,  metadata={"description": "The config key"}, validate=hconfig_key_validator)  # type: ignore
    value = StringOrBooleanField(required=True,  metadata={"description": "The config value"})


# region usage
class UsageData(Schema):
    uuid = fields.UUID(required=True,  metadata={"description": "The user uuid"})
    usage = fields.Integer(required=True,  metadata={"description": "The user usage in bytes"})
    devices = fields.List(fields.String(required=True,  metadata={"description": "The user connected devices"}))


class UsageInputOutputSchema(Schema):
    usages = fields.List(fields.Nested(UsageData), required=True,  metadata={"description": "The list of usages"})
# endregion


# region sync
class SyncInputSchema(Schema):
    domains = fields.List(fields.Nested(DomainSchema), required=False,  metadata={"description": "The list of domains"})
    proxies = fields.List(fields.Nested(ProxySchema), required=False,  metadata={"description": "The list of proxies"})
    hconfigs = fields.List(fields.Nested(HConfigSchema), required=False,  metadata={"description": "The list of configs"})
    # users = fields.List(fields.Nested(UserSchema),required=True, metadata={"description": "The list of users"})
    # admin_users = fields.List(fields.Nested(AdminSchema),required=True, metadata={"description": "The list of admin users"})

    def validate(self, data, **kwargs):
        if not (data.get("domains") or data.get("proxies") or data.get("hconfigs")):
            raise ValidationError("At least one field must exist (domains, proxies, or hconfigs)")
        return data


class SyncOutputSchema(Schema):
    users = fields.List(fields.Nested(UserSchema), required=True,  metadata={"description": "The list of users"})
    admin_users = fields.List(fields.Nested(AdminSchema), required=True,  metadata={"description": "The list of admin users"})

# endregion


# region child status
class ChildStatusInputSchema(Schema):
    child_unique_id = fields.String(required=True,  metadata={"description": "The child's unique id"})


class ChildStatusOutputSchema(Schema):
    existance = fields.Boolean(required=True,  metadata={"description": "Whether child exists"})

# end region


# region register

class RegisterDataSchema(Schema):
    users = fields.List(fields.Nested(UserSchema), required=True,  metadata={"description": "The list of users"})
    domains = fields.List(fields.Nested(DomainSchema), required=True,  metadata={"description": "The list of domains"})
    proxies = fields.List(fields.Nested(ProxySchema), required=True,  metadata={"description": "The list of proxies"})
    admin_users = fields.List(fields.Nested(AdminSchema), required=True,  metadata={"description": "The list of admin users"})
    hconfigs = fields.List(fields.Nested(HConfigSchema), required=True,  metadata={"description": "The list of configs"})


class RegisterInputSchema(Schema):
    panel_data = fields.Nested(RegisterDataSchema, required=True,  metadata={"description": "The child's data"})
    unique_id = fields.String(required=True,  metadata={"description": "The child's unique id"})
    name = fields.String(required=True,  metadata={"description": "The child's name"})
    mode = fields.Enum(ChildMode, required=True,  metadata={"description": "The child's mode"})


class RegisterOutputSchema(Schema):
    parent_unique_id = fields.String( metadata={"description": "The parent's unique id"})
    users = fields.List(fields.Nested(UserSchema), required=True,  metadata={"description": "The list of users"})
    admin_users = fields.List(fields.Nested(AdminSchema), required=True,  metadata={"description": "The list of admin users"})

# endregion



import xtlsapi
from hiddifypanel.models import *
from .abstract_driver import DriverABS
from collections import defaultdict
from hiddifypanel.cache import cache
from loguru import logger


class XrayApi(DriverABS):
    def is_enabled(self) -> bool:
        return hconfig(ConfigEnum.core_type) == "xray"

    def get_xray_client(self):
        if not hasattr(self, 'xray_client'):
            self.xray_client = xtlsapi.XrayClient('127.0.0.1', 10085)
        return self.xray_client

    def get_enabled_users(self):
        return {u:1 for u in self.get_enabled_users_terminal()}
        # xray_client = self.get_xray_client()
        # usages = xray_client.stats_query('user', reset=True)
        # res = defaultdict(int)
        # tags = set(self.get_inbound_tags())
        # for use in usages:
        #     if "user>>>" not in use.name:
        #         continue
        #     uuid = use.name.split(">>>")[1].split("@")[0]
        #     res[uuid]=1

        #     #TODO use xtls api
        #     for t in tags.copy():
        #         try:
        #             self.__add_uuid_to_tag(uuid, t)
        #             self._remove_client(uuid, [t], False)
        #             # print(f"Success add  {uuid} {t}")
        #             res[uuid] = 0
        #             break
        #         except ValueError:
        #             # tag invalid
        #             tags.remove(t)
        #             pass
        #         except xtlsapi.xtlsapi.exceptions.EmailAlreadyExists as e:
        #             res[uuid] = 1
        #             break
        #         except Exception as e:
        #             print(f"error {e}")
        #             # res[uuid] = 1
        #             break

        # return res

        # xray_client = self.get_xray_client()
        # users = User.query.all() # t = "xtls"
        # protocol = "vless"
        # enabled = {}
        # for u in users:
        #     uuid = u.uuid
        #     try:
        #         xray_client.add_client(t, f'{uuid}', f'{uuid}@hiddify.com', protocol=protocol, flow='xtls-rprx-vision', alter_id=0, cipher='chacha20_poly1305')
        #         xray_client.remove_client(t, f'{uuid}@hiddify.com')
        #         enabled[uuid] = 0
        #     except xtlsapi.xtlsapi.exceptions.EmailAlreadyExists as e:
        #         enabled[uuid] = 1
        #     except Exception as e:
        #         print(f"error {e}")
        #         enabled[uuid] = e
        # return enabled

    # @cache.cache(ttl=300)
    def get_inbound_tags(self):
        try:
            xray_client = self.get_xray_client()
            inbounds = {inb.name.split(">>>")[1] for inb in xray_client.stats_query('inbound')}
            # print(f"Success in get inbound tags {inbounds}")
        except Exception as e:
            print(f"error in get inbound tags {e}")
            inbounds = {}
        return list(inbounds)

    def __add_uuid_to_tag(self, uuid, t):
        xray_client = self.get_xray_client()
        proto_map = {
            'vless': 'vless',
            'realityin': 'vless',
            'xtls': 'vless',
            'quic': 'vless',
            'trojan': 'trojan',
            'vmess': 'vmess',
            'ss': 'shadowsocks',
            'v2ray': 'shadowsocks',
            'kcp': 'vless',
            'dispatcher': 'trojan',
            'reality': 'vless'
        }

        def proto(t):
            res = '', ''
            for p, protocol in proto_map.items():
                if p in t:
                    res = p, protocol
                    break
            return res
        p, protocol = proto(t)
        if not p:
            raise ValueError("incorrect tag")
        flow='xtls-rprx-vision' if 'realityin_tcp' in t else '\0'
        # if (protocol == "vless" and p != "xtls" and p != "realityin") or "realityingrpc" in t:
        #     xray_client.add_client(t, f'{uuid}', f'{uuid}@hiddify.com', protocol=protocol, flow='\0',)
        # else:
        xray_client.add_client(t, f'{uuid}', f'{uuid}@hiddify.com', protocol=protocol, flow=flow, alter_id=0, cipher='chacha20_poly1305')

    def add_client(self, user):
        uuid = user.uuid
        xray_client = self.get_xray_client()
        tags = self.get_inbound_tags()

        for t in tags:
            try:
                self.__add_uuid_to_tag(uuid, t)
                # print(f"Success add  {uuid} {t}")
            except ValueError:
                # tag invalid
                pass
            except Exception as e:
                # print(f"error in add  {uuid} {t} {e}")
                pass

    def remove_client(self, user):
        return self._remove_client(user.uuid)

    def _remove_client(self, uuid, tags=None, dolog=True):
        xray_client = self.get_xray_client()
        tags = tags or self.get_inbound_tags()

        for t in tags:
            try:
                xray_client.remove_client(t, f'{uuid}@hiddify.com')
                # if dolog:
                #     logger.info(f"Success remove  {uuid} {t}")
            except Exception as e:
                if dolog:
                    logger.info(f"error in remove  {uuid} {t} {e}")
                pass

    def get_all_usage(self)->dict:
        xray_client = self.get_xray_client()
        usages = xray_client.stats_query('user', reset=True)
        # uuid_user_map = {u.uuid: u for u in users}
        res = defaultdict(int)
        for use in usages:
            if "user>>>" not in use.name:
                continue
            uuid = use.name.split(">>>")[1].split("@")[0]
            # if u := uuid_user_map.get(uuid):
            res[uuid] += use.value
            # else:
            #     self._remove_client(uuid)
        return res

    def get_usage_imp(self, uuid):
        xray_client = self.get_xray_client()
        d = xray_client.get_client_download_traffic(f'{uuid}@hiddify.com', reset=True)
        u = xray_client.get_client_upload_traffic(f'{uuid}@hiddify.com', reset=True)

        res = None
        if d is None:
            res = u
        elif u is None:
            res = d
        else:
            res = d + u
        if res:
            logger.debug(f"Xray usage {uuid} d={d} u={u} sum={res}")
        return res




    def get_enabled_users_terminal(self):
        import subprocess
        import json
        tags=self.get_inbound_tags()
        for t in tags:
        # Command to execute
            cmd = [
                'xray',
                'api',
                'inbounduser',
                '--server=127.0.0.1:10085',
                f'-tag={t}'
            ]

            try:
                result = subprocess.run(cmd, capture_output=True, text=True, check=True)

                # Parse JSON output
                data = json.loads(result.stdout)
                users= [splt[0] for splt in [u.get('email','').split("@") for u in data.get('users',[])] if len(splt)==2 and splt[1]=="hiddify.com"]
                if len(data)>0:
                    return users

            except subprocess.CalledProcessError as e:
                print("Command failed:", e)
                print("Error output:", e.stderr)
            except json.JSONDecodeError as e:
                print("Failed to parse JSON:", e)
        return []



from celery import shared_task
from sqlalchemy import func
from typing import Dict
import datetime

from hiddifypanel.drivers import user_driver
from hiddifypanel.models import *
from hiddifypanel.panel import hiddify
from hiddifypanel.database import db, db_execute, text
from hiddifypanel import cache, hutils
from loguru import logger
import json
to_gig_d = 1024**3


@shared_task(ignore_result=False)
def update_local_usage():
    lock_key = "lock-update-local-usage"
    # if not cache.redis_client.set(lock_key, "locked", nx=True, ex=600):
    #     return {"msg": "last update task is not finished yet."}
    try:
        res = update_local_usage_not_lock()
        cache.redis_client.set(lock_key, "locked", nx=False, ex=60)

        return res
    except Exception as e:
        cache.redis_client.set(lock_key, "locked", nx=False, ex=60)
        logger.exception("Exception in update usage")
        raise
        return {"msg": f"Exception in update usage: {e}"}

    # return {"status": 'success', "comments":res}


def update_local_usage_not_lock():

    try:
        res = user_driver.get_users_usage(reset=True)
        return add_users_usage_new([{'uuid': uuid, "usage": uinfo['usage']} for uuid, uinfo in res.items()], child_id=0)
        # add_users_usage_uuid({"66ac79b8-8c03-4084-81c7-a2b1b3e9eefe":{"usage":1000000000}},child_id=0)
        # json_data=json.dumps([{ "uuid": uuid, "usage": uinfo["usage"]}for uuid, uinfo in {"66ac79b8-8c03-4084-81c7-a2b1b3e9eefe":{"usage":1000000000}}.items()])
        # db_execute("CALL add_usage_json(:usage_data)", usage_data= json_data, commit=True)
    except Exception as e:
        raise


def add_users_usage_uuid(uuids_bytes: Dict[str, Dict], child_id, sync=False):
    uuids_bytes = {u: v for u, v in uuids_bytes.items() if v and v.get('usage', 0) > 0}
    uuids = uuids_bytes.keys()
    users = db.session.query(User).filter(User.uuid.in_(uuids))
    dbusers_bytes = {u: uuids_bytes.get(u.uuid, {"usage": 0}) for u in users}
    _add_users_usage(dbusers_bytes, child_id, sync)  # type: ignore


def _reset_priodic_usage() -> bool:
    apply_changes = False
    last_usage_check: int = hconfig(ConfigEnum.last_priodic_usage_check) or 0
    import time
    current_time = int(time.time())
    if current_time - last_usage_check < 60 * 60 * 6:
        return apply_changes
    # reset as soon as possible in the day
    if datetime.datetime.now().hour > 5 and current_time - last_usage_check < 60 * 60 * 24:
        return apply_changes
    logger.debug("reseting user usage if needed")
    # for user in db.session.query(User).filter(User.mode != UserMode.no_reset).all():
    #     if user.user_should_reset():
    #         logger.info(f"reseting user usage for {user.uuid}")
    #         user.reset_usage(commit=False)

    today = datetime.date.today()

    db_change = False
    for user in db.session.query(User).filter(User.mode != UserMode.no_reset, User.start_date != None, User.start_date+User.package_days >= today).all():
        if user.user_should_reset():
            logger.info(f"reseting user usage for {user.uuid}")
            old_active = user.is_active
            user.reset_usage(commit=False)
            db_change = True

            if not old_active and user.is_active:
                logger.info(f"adding enabled client {user.uuid} ")
                user_driver.add_client(user)
                apply_changes = True
                send_bot_message(user)

    if db_change:
        db.session.commit()

    for user in db.session.query(User).filter(User.start_date != None, User.start_date+User.package_days < today).all():
        logger.info(f"Removing enabled client {user.uuid} ")
        if not user.is_active:
            user_driver.remove_client(user)
            apply_changes = True

    set_hconfig(ConfigEnum.last_priodic_usage_check, current_time, commit=True)
    return apply_changes


def add_users_usage_new(usages: list[dict], child_id, sync=False):
    usages = [use for use in usages if use['usage'] > 0]
    # usages[0]['usage']=1000000000000
    before_enabled_users = user_driver.get_enabled_users()

    daily_usage = {}
    cur_time=datetime.datetime.now()
    today = cur_time.date()
    db_changes = False
    for adm in db.session.query(AdminUser).all():
        daily_usage[adm.id] = db.session.query(DailyUsage).filter(DailyUsage.date == today, DailyUsage.admin_id == adm.id, DailyUsage.child_id == child_id).first()
        if daily_usage[adm.id] is None:
            logger.info(f"creating a new daily usage {today} admin={adm.id} child={child_id}")
            daily_usage[adm.id] = DailyUsage(date=today, admin_id=adm.id, child_id=child_id, usage=0)
            db.session.add(daily_usage[adm.id])
            db_changes = True
        daily_usage[adm.id].online = db.session.query(User).filter(User.added_by == adm.id).filter(func.DATE(User.last_online) == today).count()
    if db_changes:
        db.session.commit()

    apply_changes = _reset_priodic_usage()
    
    db_execute("CALL add_usage_json(:usage_data,:cur_time)", usage_data=json.dumps(usages),cur_time=cur_time.strftime('%Y-%m-%d %H:%M:%S'), commit=True)

    usage_map = {u['uuid']: u for u in usages}
    
    users = db.session.query(User).filter(User.uuid.in_(set(usage_map.keys()))).all()

    all_users_uuids = set()
    for user in users:
        all_users_uuids.add(user.uuid)

        user_before_active = before_enabled_users.get(user.uuid,False)
        user_active = user.is_active

        if not user_before_active and user_active:
            logger.info(f"Enabling disabled client {user.uuid} ")
            user_driver.add_client(user)
            send_bot_message(user)
            apply_changes = True
        elif user_before_active and not user_active:
            logger.info(f"Removing enabled client {user.uuid} ")
            user_driver.remove_client(user)
            send_bot_message(user)
            apply_changes = True

        daily_usage.get(user.added_by, daily_usage[1]).usage += usage_map[user.uuid]['usage']

    db.session.commit()

    if len(users) != len(usage_map):
        # check for zombie-users
        check_users = set(before_enabled_users.keys())-all_users_uuids
        all_db_users = {u.uuid for u in db.session.query(User).filter(User.uuid.in_(check_users)).all()}
        zombie_users = check_users-all_db_users

        for uuid in zombie_users:
            logger.info(f"Remove zombiee users {uuid} ")
            user_driver.remove_client(User(uuid=uuid))
            apply_changes = True

    if apply_changes:
        hiddify.quick_apply_users()

    return {"status": 'success', "comments": usages, "date": hutils.convert.time_to_json(cur_time)}


# def _add_users_usage(users_usage_data: Dict[User, Dict], child_id, sync=False):
#     '''
#     sync: when enabled, it means we have received usages from the parent panel
#     '''
#     res = {}
#     have_change = False
#     before_enabled_users = user_driver.get_enabled_users()
#     daily_usage = {}
#     today = datetime.date.today()
#     changes = False
#     for adm in db.session.query(AdminUser).all():
#         daily_usage[adm.id] = db.session.query(DailyUsage).filter(DailyUsage.date == today, DailyUsage.admin_id == adm.id, DailyUsage.child_id == child_id).first()
#         if daily_usage[adm.id] is None:
#             logger.info(f"creating a new daily usage {today} admin={adm.id} child={child_id}")
#             daily_usage[adm.id] = DailyUsage(date=today, admin_id=adm.id, child_id=child_id)
#             db.session.add(daily_usage[adm.id])
#             changes = True
#         daily_usage[adm.id].online = db.session.query(User).filter(User.added_by == adm.id).filter(func.DATE(User.last_online) == today).count()
#     if changes:
#         db.session.commit()
#     _reset_priodic_usage()

#     # userDetails = {p.user_id: p for p in UserDetail.query.filter(UserDetail.child_id == child_id).all()}
#     for user, uinfo in users_usage_data.items():
#         usage_bytes = uinfo['usage']

#         # UserDetails things
#         # detail = UserDetail(user_id=user.id, child_id=child_id)
#         # detail = userDetails.get(user.id)
#         # if not detail:
#         #     detail = UserDetail(user_id=user.id, child_id=child_id)
#         #     db.session.add(detail)
#         # if uinfo['devices'] != detail.connected_devices:
#         #     detail.connected_devices = uinfo['devices']

#         # Enable the user if isn't already
#         if not before_enabled_users[user.uuid] and user.is_active:
#             logger.info(f"Enabling disabled client {user.uuid} ")
#             user_driver.add_client(user)
#             send_bot_message(user)
#             have_change = True

#         # Check if there's new usage value
#         if not isinstance(usage_bytes, int) or usage_bytes == 0:
#             res[user.uuid] = "No usage"
#         else:
#             # Set new daily usage of the user
#             if sync and daily_usage.get(user.added_by, daily_usage[1]).usage != usage_bytes:
#                 daily_usage.get(user.added_by, daily_usage[1]).usage = usage_bytes
#             else:
#                 daily_usage.get(user.added_by, daily_usage[1]).usage += usage_bytes

#             # Set new current usage of the user
#             if sync and user.current_usage != usage_bytes:
#                 user.current_usage = usage_bytes
#                 # detail.current_usage_GB = in_gig
#             else:
#                 user.current_usage += usage_bytes
#                 # detail.current_usage = detail.current_usage or 0
#                 # detail.current_usage += usage_bytes

#             # Change last online time of the user
#             user.last_online = datetime.datetime.now()
#             # detail.last_online = datetime.datetime.now()

#             # Set start date of user to the current datetime if it hasn't been set already
#             if user.start_date is None:
#                 user.start_date = datetime.date.today()

#             res[user.uuid] = f'{usage_bytes/1000000:0.3f}MB'

#         # Remove user from drivers(singbox, xray, wireguard etc.) if they're inactive
#         # print(before_enabled_users[user.uuid], user.is_active)
#         if before_enabled_users[user.uuid] and not user.is_active:
#             logger.info(f"Removing enabled client {user.uuid} ")

#             user_driver.remove_client(user)
#             have_change = True
#             res[user.uuid] = f"{res[user.uuid]} !OUT of USAGE! Client Removed"

#     db.session.commit()  # type: ignore

#     # Remove invalid users
#     for uuid in before_enabled_users:
#         if uuid in res:
#             continue

#         user = db.session.query(User).filter(User.uuid == uuid).first()
#         if not user:
#             user_driver.remove_client(User(uuid=uuid))
#         elif not user.is_active:
#             user_driver.remove_client(user)

#     # print("------------------", res)
#     # Apply the changes to the drivers
#     if have_change:
#         hiddify.quick_apply_users()

#     # Sync the new data with the parent node if the data has not been set by the parent node itself and the current panel is a child panel
#     if not sync and hutils.node.is_child():
#         hutils.node.child.sync_users_usage_with_parent()

#     return {"status": 'success', "comments": res, "date": hutils.convert.time_to_json(datetime.datetime.now())}


def send_bot_message(user):
    if not (hconfig(ConfigEnum.telegram_bot_token) or hutils.node.is_child()):
        return
    if not user.telegram_id:
        return
    from flask_babel import lazy_gettext as _
    from hiddifypanel.panel.commercial.telegrambot import bot, Usage
    try:
        msg = Usage.get_usage_msg(user.uuid)
        msg = _("User activated!") if user.is_active else _("Package ended!") + "\n" + msg
        bot.send_message(user.telegram_id, msg, reply_markup=Usage.user_keyboard(user.uuid))
    except BaseException:
        pass



from hiddifypanel.models import *
from hiddifypanel.panel.admin.adminlte import AdminLTEModelView
from flask_babel import gettext as __
from flask_babel import lazy_gettext as _
from flask import g, redirect
from markupsafe import Markup
from hiddifypanel.auth import login_required
from flask_admin.actions import action
from flask_admin.contrib.sqla import form, filters as sqla_filters, tools
from flask_admin import expose
from hiddifypanel.panel import  custom_widgets

# Define a custom field type for the related domains
from hiddifypanel import hutils

from wtforms.widgets import TextArea
import json

class ProxyDetailsAdmin(AdminLTEModelView):
    list_template = 'model/proxydetail_list.html'
    column_hide_backrefs = True
    can_create = False
    form_excluded_columns = ['child', 'proto', 'transport', 'cdn']
    column_exclude_list = ['child']
    column_searchable_list = ['name', 'proto', 'transport', 'l3', 'cdn']
    column_editable_list = ['name']
    form_extra_fields = {
        'params': custom_widgets.JSONField(label='Proxy Params')
    }
    form_overrides={
          'params': custom_widgets.JSONField
    }
    @expose('reset_proxies')
    def reset_proxies(self):
        from hiddifypanel.panel.init_db import get_proxy_rows_v1
        from hiddifypanel.database import db
        db.session.bulk_save_objects(get_proxy_rows_v1())
        db.session.commit()
        hutils.flask.flash((_('config.validation-success-no-reset')), 'success')  # type: ignore
        return redirect("./")

    @action('disable', 'Disable', 'Are you sure you want to disable selected proxies?')
    def action_disable(self, ids):
        query = tools.get_query_for_ids(self.get_query(), self.model, ids)
        count = query.update({'enable': False})

        self.session.commit()
        hutils.flask.flash(_('%(count)s records were successfully disabled.', count=count), 'success')
        hutils.proxy.get_proxies.invalidate_all()

    @action('enable', 'Enable', 'Are you sure you want to enable selected proxies?')
    def action_enable(self, ids):
        query = tools.get_query_for_ids(self.get_query(), self.model, ids)
        count = query.update({'enable': True})

        self.session.commit()
        hutils.flask.flash(_('%(count)s records were successfully enabled.', count=count), 'success')
        hutils.proxy.get_proxies.invalidate_all()

    # list_template = 'model/domain_list.html'

    # form_overrides = {'work_with': Select2Field}

    def after_model_change(self, form, model, is_created):
        if hutils.node.is_child():
            hutils.node.run_node_op_in_bg(hutils.node.child.sync_with_parent, *[hutils.node.child.SyncFields.proxies])
        hutils.proxy.get_proxies.invalidate_all()
        pass

    def after_model_delete(self, model):
        if hutils.node.is_child():
            hutils.node.run_node_op_in_bg(hutils.node.child.sync_with_parent, *[hutils.node.child.SyncFields.proxies])
        hutils.proxy.get_proxies.invalidate_all()
        pass

    def is_accessible(self):
        if login_required(roles={Role.super_admin, Role.admin})(lambda: True)() != True:
            return False
        return True
    def _params_formatter(view, context, model, name):
        return str(model.params)
    def _enable_formatter(view, context, model, name):
        if model.enable:
            link = '<i class="fa-solid fa-circle-check text-success"></i> '
        else:
            link = '<i class="fa-solid fa-circle-xmark text-danger"></i> '
        return Markup(link)
    column_formatters = {

        "enable": _enable_formatter,
        "params": _params_formatter
    }



import datetime
import uuid
from hiddifypanel.models.role import Role
from sqlalchemy import Column, String, BigInteger, Enum

from flask_login import UserMixin as FlaskLoginUserMixin
from hiddifypanel.models import Lang
from hiddifypanel.database import db



class BaseAccount(db.Model, FlaskLoginUserMixin):  # type: ignore
    __abstract__ = True
    uuid = Column(String(36), default=lambda: str(uuid.uuid4()), nullable=False, unique=True, index=True)
    name = Column(String(512), nullable=False, default='')
    username = Column(String(100), nullable=True, default='', index=True)
    password = Column(String(100), nullable=True, default='')
    comment = Column(String(512), nullable=True, default='')
    telegram_id = Column(BigInteger, nullable=True, default=None, index=True)
    lang = Column(Enum(Lang), default=None)

    @property
    def role(self) -> Role | None:
        return None

    def get_id(self) -> str | None:
        return f'{self.__class__.name}_{self.id if self.hasattr("id") else "-"}'

    def is_username_unique(self) -> bool:
        cls = self.__class__()
        model = cls.query.filter(cls.username == self.username, cls.id != self.id).first()
        if model:
            return False
        return True

    def to_dict(self, convert_date=True) -> dict:
        return {
            'name': self.name,
            'comment': self.comment,
            'uuid': self.uuid,
            'telegram_id': self.telegram_id,
            'lang': self.lang
        }
    def update_password(self,new_password):
        self.password=new_password
        db.session.commit()

    @classmethod
    def by_id(cls, id: int):
        # return cls.query.filter(cls.id == id).first()
        return db.session.query(cls).get(id)

    @classmethod
    def by_uuid(cls, uuid: str, create: bool = False):
        if not isinstance(uuid, str):
            uuid = str(uuid)
        account = cls.query.filter(cls.uuid == uuid).first()
        if not account and create:
            raise NotImplementedError
        return account

    @classmethod
    def by_username_password(cls, username: str, password: str):
        return cls.query.filter(cls.username == username, cls.password == password).first()

    @classmethod
    def add_or_update(cls, commit: bool = True, old_uuid=None, **data):
        db_account: BaseAccount = cls.by_uuid(old_uuid or data.get('uuid'), create=True)
        from hiddifypanel import hutils
        if hutils.auth.is_uuid_valid(data.get('uuid')):
            db_account.uuid = data['uuid']

        if data.get('name') is not None:
            db_account.name = data.get('name')

        if data.get('comment') is not None:
            db_account.comment = data.get('comment')
        if data.get('telegram_id') is not None:
            db_account.telegram_id = hutils.convert.to_int(data.get('telegram_id'))
        if data.get('lang') is not None:
            db_account.lang = data.get('lang')
        if commit:
            db.session.commit()  # type: ignore
        return db_account

    @classmethod
    def bulk_register(cls, accounts: list = [], commit: bool = True, remove: bool = False):
        for u in accounts:
            cls.add_or_update(commit=False, **u)
        if remove:
            dd = {str(u['uuid']): 1 for u in accounts}
            for d in cls.query.all():
                if d.uuid not in dd:
                    db.session.delete(d)  # type: ignore
        if commit:
            db.session.commit()  # type: ignore



from enum import auto, Enum
import os
from typing import Union

from strenum import StrEnum
from fast_enum import FastEnum


class HEnum(StrEnum):
    @classmethod
    def from_str(cls, key: str) -> 'HEnum':
        return cls[key]


class Lang(HEnum):
    en = auto()
    fa = auto()
    ru = auto()
    pt = auto()
    zh = auto()
    my = auto()


class PanelMode(HEnum):
    standalone = auto()
    parent = auto()
    child = auto()


class MieruMultiplexing(HEnum):
    MULTIPLEXING_DEFAULT = auto()
    MULTIPLEXING_LOW = auto()
    MULTIPLEXING_MIDDLE = auto()
    MULTIPLEXING_HIGH = auto()

class MieruHandshake(HEnum):
    HANDSHAKE_DEFAULT = auto()
    HANDSHAKE_NO_WAIT = auto()
    HANDSHAKE_STANDARD = auto()
    

class LogLevel(HEnum):
    TRACE = auto()
    DEBUG = auto()
    INFO = auto()
    SUCCESS = auto()
    WARNING = auto()
    ERROR = auto()
    CRITICAL = auto()


class ConfigCategory(StrEnum):
    admin = auto()
    branding = auto()
    general = auto()
    proxies = auto()
    domain_fronting = auto()
    telegram = auto()
    http = auto()
    tls = auto()
    mux = auto()
    tls_trick = auto()
    ssh = auto()
    ssfaketls = auto()
    mieru = auto()
    shadowtls = auto()
    restls = auto()
    tuic = auto()
    hysteria = auto()
    ssr = auto()
    kcp = auto()
    hidden = auto()
    advanced = auto()
    too_advanced = auto()
    warp = auto()
    reality = auto()
    wireguard = auto()
    shadowsocks = auto()


class ApplyMode(StrEnum):
    apply_config = auto()
    reinstall = auto()
    nothing = auto()


def _BoolConfigDscr(category: ConfigCategory, apply_mode: ApplyMode = ApplyMode.nothing, show_in_parent: bool = True, hide_in_virtual_child=False) -> "ConfigEnum":
    return category, apply_mode, bool, show_in_parent


def _StrConfigDscr(category: ConfigCategory, apply_mode: ApplyMode = ApplyMode.nothing, show_in_parent: bool = True, hide_in_virtual_child=False) -> "ConfigEnum":
    return category, apply_mode, str, show_in_parent


def _IntConfigDscr(category: ConfigCategory, apply_mode: ApplyMode = ApplyMode.nothing, show_in_parent: bool = True, hide_in_virtual_child=False) -> "ConfigEnum":
    return category, apply_mode, int, show_in_parent


def _TypedConfigDscr(ctype: type, category: ConfigCategory, apply_mode: ApplyMode = ApplyMode.nothing, show_in_parent: bool = True, hide_in_virtual_child=False) -> "ConfigEnum":
    return category, apply_mode, ctype, show_in_parent


class ConfigEnum(metaclass=FastEnum):
    # category: ConfigCategory
    __slots__ = ('name', 'value', 'category', 'apply_mode', 'type', 'show_in_parent', 'hide_in_virtual_child')

    def __init__(self, category: ConfigCategory, apply_mode: ApplyMode = ApplyMode.apply_config, ctype=type, show_in_parent: bool = True, hide_in_virtual_child=False, name=auto):
        self.value = name
        self.name = name
        self.category = category
        self.apply_mode = apply_mode
        self.type = ctype
        self.show_in_parent = show_in_parent
        self.hide_in_virtual_child = hide_in_virtual_child

    @classmethod
    def dbvalues(cls):
        return {c.name: c for c in ConfigEnum}
    create_easysetup_link = _BoolConfigDscr(ConfigCategory.hidden, ApplyMode.nothing, hide_in_virtual_child=True)
    wireguard_enable = _BoolConfigDscr(ConfigCategory.wireguard, ApplyMode.reinstall, hide_in_virtual_child=True)
    wireguard_port = _StrConfigDscr(ConfigCategory.wireguard, ApplyMode.apply_config, hide_in_virtual_child=True)
    wireguard_ipv6 = _StrConfigDscr(ConfigCategory.hidden, ApplyMode.apply_config, hide_in_virtual_child=True)
    wireguard_ipv4 = _StrConfigDscr(ConfigCategory.hidden, ApplyMode.apply_config, hide_in_virtual_child=True)
    wireguard_private_key = _StrConfigDscr(ConfigCategory.hidden, ApplyMode.apply_config, hide_in_virtual_child=True)
    wireguard_public_key = _StrConfigDscr(ConfigCategory.hidden, ApplyMode.apply_config, hide_in_virtual_child=True)
    wireguard_noise_trick = _StrConfigDscr(ConfigCategory.wireguard, ApplyMode.apply_config)

    ssh_server_redis_url = _StrConfigDscr(ConfigCategory.hidden, hide_in_virtual_child=True)
    ssh_server_port = _StrConfigDscr(ConfigCategory.ssh, ApplyMode.apply_config, hide_in_virtual_child=True)
    ssh_server_enable = _BoolConfigDscr(ConfigCategory.ssh, ApplyMode.reinstall)
    first_setup = _BoolConfigDscr(ConfigCategory.hidden)
    core_type = _StrConfigDscr(ConfigCategory.advanced, ApplyMode.reinstall, hide_in_virtual_child=True)
    warp_enable = _BoolConfigDscr(ConfigCategory.hidden, ApplyMode.reinstall, hide_in_virtual_child=True)
    warp_mode = _StrConfigDscr(ConfigCategory.warp, ApplyMode.apply_config, hide_in_virtual_child=True)
    warp_plus_code = _StrConfigDscr(ConfigCategory.warp, ApplyMode.apply_config, hide_in_virtual_child=True)
    warp_sites = _StrConfigDscr(ConfigCategory.warp, ApplyMode.apply_config, hide_in_virtual_child=True)
    dns_server = _StrConfigDscr(ConfigCategory.general, ApplyMode.apply_config, hide_in_virtual_child=True)
    reality_fallback_domain = _StrConfigDscr(ConfigCategory.hidden, ApplyMode.apply_config)  # removed
    reality_server_names = _StrConfigDscr(ConfigCategory.hidden, ApplyMode.apply_config)  # removed
    reality_short_ids = _StrConfigDscr(ConfigCategory.reality, ApplyMode.apply_config, hide_in_virtual_child=True)
    reality_private_key = _StrConfigDscr(ConfigCategory.reality, ApplyMode.apply_config, hide_in_virtual_child=True)
    reality_public_key = _StrConfigDscr(ConfigCategory.reality, ApplyMode.apply_config, hide_in_virtual_child=True)
    reality_port = _StrConfigDscr(ConfigCategory.hidden, ApplyMode.apply_config, hide_in_virtual_child=True)
    special_port = _StrConfigDscr(ConfigCategory.reality, ApplyMode.apply_config, hide_in_virtual_child=True)
    

    restls1_2_domain = _StrConfigDscr(ConfigCategory.hidden)
    restls1_3_domain = _StrConfigDscr(ConfigCategory.hidden)
    show_usage_in_sublink = _BoolConfigDscr(ConfigCategory.general)
    cloudflare = _StrConfigDscr(ConfigCategory.too_advanced)
    license = _StrConfigDscr(ConfigCategory.hidden)
    country = _StrConfigDscr(ConfigCategory.general, ApplyMode.reinstall, hide_in_virtual_child=True)
    package_mode = _StrConfigDscr(ConfigCategory.advanced, hide_in_virtual_child=True)
    utls = _StrConfigDscr(ConfigCategory.advanced)
    telegram_bot_token = _StrConfigDscr(ConfigCategory.telegram, hide_in_virtual_child=True)

    # region child-parent
    # deprecated
    is_parent = _BoolConfigDscr(ConfigCategory.hidden)
    # parent panel domain
    parent_panel = _StrConfigDscr(ConfigCategory.hidden)  # should be able to change by user
    parent_domain = _StrConfigDscr(ConfigCategory.hidden)
    parent_admin_proxy_path = _StrConfigDscr(ConfigCategory.hidden)

    # the panel mode could be one of these: "parent", "child", "standalone"
    # this config value would be 'standalone' by default. and would be set by panel itself
    panel_mode = _TypedConfigDscr(PanelMode, ConfigCategory.hidden, hide_in_virtual_child=True)
    # endregion

    log_level = _TypedConfigDscr(LogLevel, ConfigCategory.hidden, ApplyMode.reinstall, hide_in_virtual_child=True)

    unique_id = _StrConfigDscr(ConfigCategory.hidden)
    last_hash = _StrConfigDscr(ConfigCategory.hidden)
    cdn_forced_host = _StrConfigDscr(ConfigCategory.hidden)  # removed
    lang = _TypedConfigDscr(Lang, ConfigCategory.branding)
    admin_lang = _TypedConfigDscr(Lang, ConfigCategory.admin)
    admin_secret = _StrConfigDscr(ConfigCategory.hidden)  # removed

    default_useragent_string = _StrConfigDscr(ConfigCategory.general)    
    use_ip_in_config=_BoolConfigDscr(ConfigCategory.hidden)
    # tls
    tls_ports = _StrConfigDscr(ConfigCategory.tls, ApplyMode.apply_config)

    tls_fragment_enable = _BoolConfigDscr(ConfigCategory.tls_trick)
    tls_fragment_size = _StrConfigDscr(ConfigCategory.tls_trick)
    tls_fragment_sleep = _StrConfigDscr(ConfigCategory.tls_trick)
    tls_fragment_packets = _StrConfigDscr(ConfigCategory.tls_trick)
    tls_mixed_case = _BoolConfigDscr(ConfigCategory.tls_trick)
    tls_padding_enable = _BoolConfigDscr(ConfigCategory.tls_trick, ApplyMode.apply_config)
    tls_padding_length = _StrConfigDscr(ConfigCategory.tls_trick, ApplyMode.apply_config)
    tls_ech_enable = _BoolConfigDscr(ConfigCategory.tls, ApplyMode.apply_config)
    

    # mux
    mux_enable = _BoolConfigDscr(ConfigCategory.mux, ApplyMode.apply_config)
    mux_protocol = _StrConfigDscr(ConfigCategory.mux, ApplyMode.apply_config)
    mux_max_connections = _IntConfigDscr(ConfigCategory.mux, ApplyMode.apply_config)
    mux_min_streams = _IntConfigDscr(ConfigCategory.mux, ApplyMode.apply_config)
    mux_max_streams = _IntConfigDscr(ConfigCategory.mux, ApplyMode.apply_config)
    mux_padding_enable = _BoolConfigDscr(ConfigCategory.mux, ApplyMode.apply_config)
    mux_brutal_enable = _BoolConfigDscr(ConfigCategory.mux, ApplyMode.apply_config)
    mux_brutal_up_mbps = _IntConfigDscr(ConfigCategory.mux, ApplyMode.apply_config)
    mux_brutal_down_mbps = _IntConfigDscr(ConfigCategory.mux, ApplyMode.apply_config)

    http_ports = _StrConfigDscr(ConfigCategory.http, ApplyMode.apply_config)
    mieru_tcp_ports = _StrConfigDscr(ConfigCategory.mieru, ApplyMode.apply_config, hide_in_virtual_child=True)
    mieru_udp_ports = _StrConfigDscr(ConfigCategory.mieru, ApplyMode.apply_config, hide_in_virtual_child=True)
    kcp_ports = _StrConfigDscr(ConfigCategory.hidden, ApplyMode.apply_config)
    kcp_enable = _BoolConfigDscr(ConfigCategory.hidden, ApplyMode.apply_config)
    decoy_domain = _StrConfigDscr(ConfigCategory.general, ApplyMode.apply_config, hide_in_virtual_child=True)

    dnstt_enable = _BoolConfigDscr(ConfigCategory.proxies, ApplyMode.apply_config, hide_in_virtual_child=True)
    dnstt_resolvers = _StrConfigDscr(ConfigCategory.proxies, ApplyMode.apply_config, hide_in_virtual_child=True)

    # will be deprecated
    proxy_path = _StrConfigDscr(ConfigCategory.hidden, ApplyMode.apply_config, hide_in_virtual_child=True)
    proxy_path_admin = _StrConfigDscr(ConfigCategory.too_advanced, ApplyMode.apply_config, hide_in_virtual_child=True)
    proxy_path_client = _StrConfigDscr(ConfigCategory.too_advanced, ApplyMode.apply_config, hide_in_virtual_child=True)
    firewall = _BoolConfigDscr(ConfigCategory.general, ApplyMode.apply_config, hide_in_virtual_child=True)
    netdata = _BoolConfigDscr(ConfigCategory.hidden, ApplyMode.reinstall)  # removed
    http_proxy_enable = _BoolConfigDscr(ConfigCategory.http)
    block_iran_sites = _BoolConfigDscr(ConfigCategory.proxies, ApplyMode.apply_config, hide_in_virtual_child=True)
    allow_invalid_sni = _BoolConfigDscr(ConfigCategory.tls, ApplyMode.apply_config, hide_in_virtual_child=True)
    auto_update = _BoolConfigDscr(ConfigCategory.hidden if os.environ.get('HIDDIFY_DISABLE_UPDATE',"").lower() in {'1',"true"} else ConfigCategory.general, ApplyMode.apply_config, True, hide_in_virtual_child=True)
    speed_test = _BoolConfigDscr(ConfigCategory.general, ApplyMode.reinstall, hide_in_virtual_child=True)
    only_ipv4 = _BoolConfigDscr(ConfigCategory.general, ApplyMode.apply_config, hide_in_virtual_child=True)

    shared_secret = _StrConfigDscr(ConfigCategory.proxies, ApplyMode.apply_config, hide_in_virtual_child=True)

    telegram_enable = _BoolConfigDscr(ConfigCategory.telegram, ApplyMode.reinstall)
    # telegram_secret=auto()
    telegram_adtag = _StrConfigDscr(ConfigCategory.telegram, ApplyMode.reinstall, hide_in_virtual_child=True)
    telegram_lib = _StrConfigDscr(ConfigCategory.telegram, ApplyMode.reinstall, hide_in_virtual_child=True)
    telegram_fakedomain = _StrConfigDscr(ConfigCategory.telegram, ApplyMode.reinstall, hide_in_virtual_child=True)

    v2ray_enable = _BoolConfigDscr(ConfigCategory.hidden, ApplyMode.reinstall)
    torrent_block = _BoolConfigDscr(ConfigCategory.general, ApplyMode.apply_config)

    tuic_enable = _BoolConfigDscr(ConfigCategory.tuic, ApplyMode.apply_config)
    tuic_port = _StrConfigDscr(ConfigCategory.tuic, ApplyMode.apply_config, hide_in_virtual_child=True)

    # the hysteria is refereing to hysteria2
    hysteria_enable = _BoolConfigDscr(ConfigCategory.hysteria, ApplyMode.apply_config)
    hysteria_port = _StrConfigDscr(ConfigCategory.hidden, ApplyMode.apply_config, hide_in_virtual_child=True)
    # if be enable hysteria2 will be use salamander as obfs
    hysteria_obfs_enable = _BoolConfigDscr(ConfigCategory.hysteria, ApplyMode.apply_config)
    hysteria_up_mbps = _StrConfigDscr(ConfigCategory.hysteria, ApplyMode.apply_config)
    hysteria_down_mbps = _StrConfigDscr(ConfigCategory.hysteria, ApplyMode.apply_config)

    shadowsocks2022_enable = _BoolConfigDscr(ConfigCategory.shadowsocks, ApplyMode.apply_config)
    shadowsocks2022_method = _StrConfigDscr(ConfigCategory.hidden, ApplyMode.apply_config)
    shadowsocks2022_port = _StrConfigDscr(ConfigCategory.shadowsocks, ApplyMode.apply_config)
    ssfaketls_enable = _BoolConfigDscr(ConfigCategory.shadowsocks, ApplyMode.reinstall)
    ssfaketls_fakedomain = _StrConfigDscr(ConfigCategory.shadowsocks, ApplyMode.apply_config, hide_in_virtual_child=True)
    shadowtls_enable = _BoolConfigDscr(ConfigCategory.shadowsocks, ApplyMode.apply_config)
    shadowtls_fakedomain = _StrConfigDscr(ConfigCategory.shadowsocks, ApplyMode.apply_config, hide_in_virtual_child=True)

    ssr_enable = _BoolConfigDscr(ConfigCategory.hidden, ApplyMode.apply_config)
    # ssr_secret="ssr_secret"
    ssr_fakedomain = _StrConfigDscr(ConfigCategory.hidden, ApplyMode.apply_config)

    vmess_enable = _BoolConfigDscr(ConfigCategory.proxies, ApplyMode.apply_config)
    domain_fronting_domain = _StrConfigDscr(ConfigCategory.hidden)  # removed
    domain_fronting_http_enable = _BoolConfigDscr(ConfigCategory.hidden, ApplyMode.apply_config)  # removed
    domain_fronting_tls_enable = _BoolConfigDscr(ConfigCategory.hidden, ApplyMode.apply_config)  # removed

    ws_enable = _BoolConfigDscr(ConfigCategory.proxies, ApplyMode.apply_config)
    grpc_enable = _BoolConfigDscr(ConfigCategory.proxies, ApplyMode.apply_config)
    httpupgrade_enable = _BoolConfigDscr(ConfigCategory.proxies, ApplyMode.apply_config)
    xhttp_enable = _BoolConfigDscr(ConfigCategory.proxies, ApplyMode.apply_config)

    naive_enable = _BoolConfigDscr(ConfigCategory.proxies, ApplyMode.apply_config)
    naive_port = _StrConfigDscr(ConfigCategory.proxies, ApplyMode.apply_config)
    mieru_enable = _BoolConfigDscr(ConfigCategory.mieru, ApplyMode.apply_config)
    mieru_multiplexing =_TypedConfigDscr(MieruMultiplexing, ConfigCategory.mieru)
    mieru_handshake =_TypedConfigDscr(MieruHandshake, ConfigCategory.mieru)
    vless_enable = _BoolConfigDscr(ConfigCategory.proxies, ApplyMode.apply_config)
    trojan_enable = _BoolConfigDscr(ConfigCategory.proxies, ApplyMode.apply_config)
    reality_enable = _BoolConfigDscr(ConfigCategory.proxies, ApplyMode.apply_config)
    tcp_enable = _BoolConfigDscr(ConfigCategory.proxies, ApplyMode.apply_config)
    quic_enable = _BoolConfigDscr(ConfigCategory.proxies, ApplyMode.apply_config)
    
    xtls_enable = _BoolConfigDscr(ConfigCategory.hidden, ApplyMode.apply_config)#deprecated 
    h2_enable = _BoolConfigDscr(ConfigCategory.proxies, ApplyMode.apply_config)#deprecated

    db_version = _StrConfigDscr(ConfigCategory.hidden)
    last_priodic_usage_check = _IntConfigDscr(ConfigCategory.hidden)

    branding_title = _StrConfigDscr(ConfigCategory.branding)
    branding_site = _StrConfigDscr(ConfigCategory.branding)
    branding_freetext = _StrConfigDscr(ConfigCategory.branding)
    not_found = _StrConfigDscr(ConfigCategory.hidden)
    path_vmess = _StrConfigDscr(ConfigCategory.too_advanced, ApplyMode.apply_config, hide_in_virtual_child=True)
    path_vless = _StrConfigDscr(ConfigCategory.too_advanced, ApplyMode.apply_config, hide_in_virtual_child=True)
    path_trojan = _StrConfigDscr(ConfigCategory.too_advanced, ApplyMode.apply_config, hide_in_virtual_child=True)
    path_naive = _StrConfigDscr(ConfigCategory.too_advanced, ApplyMode.apply_config, hide_in_virtual_child=True)
    path_v2ray = _StrConfigDscr(ConfigCategory.hidden, ApplyMode.apply_config, hide_in_virtual_child=True)  # deprecated
    path_ss = _StrConfigDscr(ConfigCategory.hidden, ApplyMode.apply_config, hide_in_virtual_child=True)

    path_xhttp = _StrConfigDscr(ConfigCategory.too_advanced, ApplyMode.apply_config, hide_in_virtual_child=True)
    path_httpupgrade = _StrConfigDscr(ConfigCategory.too_advanced, ApplyMode.apply_config, hide_in_virtual_child=True)
    path_ws = _StrConfigDscr(ConfigCategory.too_advanced, ApplyMode.apply_config, hide_in_virtual_child=True)
    path_tcp = _StrConfigDscr(ConfigCategory.too_advanced, ApplyMode.apply_config, hide_in_virtual_child=True)
    path_grpc = _StrConfigDscr(ConfigCategory.too_advanced, ApplyMode.apply_config, hide_in_virtual_child=True)

    # subs
    sub_full_singbox_enable = _BoolConfigDscr(ConfigCategory.hidden)
    sub_singbox_ssh_enable = _BoolConfigDscr(ConfigCategory.hidden)
    sub_full_xray_json_enable = _BoolConfigDscr(ConfigCategory.proxies)
    sub_full_links_enable = _BoolConfigDscr(ConfigCategory.hidden)
    sub_full_links_b64_enable = _BoolConfigDscr(ConfigCategory.hidden)
    sub_full_clash_enable = _BoolConfigDscr(ConfigCategory.hidden)
    sub_full_clash_meta_enable = _BoolConfigDscr(ConfigCategory.hidden)


    #ssh host keys
    ssh_host_rsa_pk = _StrConfigDscr(ConfigCategory.hidden)
    ssh_host_rsa_pub = _StrConfigDscr(ConfigCategory.hidden)
    ssh_host_ed25519_pk = _StrConfigDscr(ConfigCategory.hidden)
    ssh_host_ed25519_pub = _StrConfigDscr(ConfigCategory.hidden)
    ssh_host_ecdsa_pk = _StrConfigDscr(ConfigCategory.hidden)
    ssh_host_ecdsa_pub = _StrConfigDscr(ConfigCategory.hidden)
    ssh_host_dsa_pk = _StrConfigDscr(ConfigCategory.hidden)
    ssh_host_dsa_pub = _StrConfigDscr(ConfigCategory.hidden)
    
    
    

    hiddifycli_enable = _BoolConfigDscr(ConfigCategory.hidden, ApplyMode.reinstall)

    @classmethod
    def __missing__(cls, value):
        return ConfigEnum.not_found

    def __contains__(self, other):
        return other in self.name

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return f'{self}' == f'{other}'

    def __neg__(self, other):
        return not self.__eq__(other)

    def endswith(self, other):
        return self.name.endswith(other)  # type: ignore

    def startswith(self, other):
        return self.name.startswith(other)  # type: ignore



from . import bot


@bot.message_handler(func=lambda message: True)
def not_handled(message):
    bot.reply_to(message, "We can not understand your request")


@bot.callback_query_handler(func=lambda call: True)
def update_usage_callback(call):  # <- passes a CallbackQuery type object to your function
    bot.answer_callback_query(call.id, "Incorrect callback", cache_time=1)
